#!/bin/sh

NotifyError () {
	# 2 means show error icon
	echo "2$1" >> /var/apresa/data/to_say
	echo "$1" >> $LOGFILE
}

InstallUpdate () {
	echo "`date` Installing update" >> $LOGFILE
	echo "`date` Update File: $1" >> $LOGFILE

	# Perform checks
	if [ -z $1 ]
	then
		NotifyError "Update file not specified"
		return 1;
	fi
	if [ ! -f $1 ]
	then
		NotifyError "Update file does not exist"
		return 1;
	fi
	if ! tar tzf "$1"
	then
		NotifyError "Update file is not a valid TGZ file"
		return 1;
	fi

	# Remove previous version of update directory
	# We want a clean directory to start with
	rm -r /tmp/update
	mkdir /tmp/update
	# Extract update file in temp update dir
	cd /tmp/update
	tar xzf $1
	if [ "$?" -ne "0" ];
	then
		NotifyError "Update file is not a valid TGZ file"
		return;
	fi

	if [ -f /var/apresa/conf/unofficial_updates ];
	then
		echo "No signature verification" >> $LOGFILE
	else
		# verify signatures
		gpg --no-default-keyring --keyring /usr/share/apresa/vididev.pubkey --verify update.sig update-r.tgz
		if [ "$?" -ne "0" ];
		then
			NotifyError "Update signature failed and/or absent (tgz)"
			return 1;
		fi
		gpg --no-default-keyring --keyring /usr/share/apresa/vididev.pubkey --verify install.sig install
		if [ "$?" -ne "0" ];
		then
			NotifyError "Update signature failed and/or absent (install)"
			return 1;
		fi
	fi
	# echo "0Verified update signature" >> /var/apresa/data/to_say

	# signatures are verified
	# it is now save to run update script

  echo "`date` Stop before update" >> $LOGFILE
	# Stop what we want to update
	service orkaudio stop	
  systemctl stop apresa_handler.socket
  service apresa_handler stop
  service apresafifo stop
	service apresacrec stop
	service apresaremote stop
	service apresaav32 stop
	service apresaavnet stop
  
  echo "`date` Run install script" >> $LOGFILE
	./install >> $LOGFILE

	echo "`date` install finished" >> $LOGFILE
	echo "0~System.UpdateFinished" >> /var/apresa/data/to_say
  return 0
}

CheckEnabled () {
  if [ ! -f /etc/init.d/$1 ] && [ ! -f /etc/systemd/system/$1.service ]; then
    echo "$1 is not installed" >> $LOGFILE
    return 0
  fi
  
  systemctl is-enabled $1
  if [ "$?" -ne "0" ];
  then
    echo "$1 is not enabled, try to enabled it" >> $LOGFILE
    systemctl enable $1
    if [ "$?" -ne "0" ];
    then
      echo "$1 could not be enabled" >> $LOGFILE
    else
      echo "$1 is enabled" >> $LOGFILE
    fi
  else
      echo "$1 is already enabled" >> $LOGFILE
  fi

  return 0
}

# here function def ends and normal execution code begins

LOGFILE=/var/log/apresa/system.log

echo "`date` Update script begin" >> $LOGFILE

if [ -z "$1" ]
then
  echo "No update file specified" >> $LOGFILE
  exit
fi

if [ -f /var/apresa/data/busy ]
then
  NotifyError "The update is not installed, because another system action is busy."
  exit
fi

echo "update" > /var/apresa/data/busy

if [ -f /var/apresa/conf/opt_cluster ]
then
	# save original maintenace mode
	ORI_CLUSTER_MAINTENANCE_MODE=0
	crm configure show cib-bootstrap-options | grep "maintenance-mode=on"
	if [ "$?" -eq "0" ]
	then
		ORI_CLUSTER_MAINTENANCE_MODE=1
	fi
	# put cluster in maintence mode
	crm configure property maintenance-mode=on
fi

InstallUpdate $1
precheckfail=$?

# clean up

# Remove update directory
rm -r /tmp/update
# Remove update file
rm $1

if [ "$precheckfail" -eq "0" ]
then
  # update installed

  # All services should be enabled at this point during upgrade to newer version
  # but downgrade can leave some services not enabled, so they are checked here
  CheckEnabled orkaudio
  CheckEnabled apresa_handler
  CheckEnabled apresacrec

  echo "`date` Start again" >> $LOGFILE
  # Start again
  /bin/sh /usr/share/apresa/apresa_startup.sh -o
  service orkaudio start
  service apresa_handler start
  service apresafifo start
  if [ -d /usr/local/lib/shcti ]; then service apresacrec start; fi
  service apresaremote start
  service apresaav32 start
  service apresaavnet start
  
else
  echo "`date` Update not performed" >> $LOGFILE
fi  

if [ -f /var/apresa/conf/opt_cluster ]
then
	# restore original cluster maintenace mode
	if [ "$ORI_CLUSTER_MAINTENANCE_MODE" -eq "0" ]
	then
		crm configure property maintenance-mode=off
	fi
fi

rm /var/apresa/data/busy
echo "`date` Update script end" >> $LOGFILE
