#!/bin/bash

# Function defs

function norecover()
{
	echo "Recovery cancelled"
	exit 1
}

function printusage()
{
	echo "Usage: recoverdb.sh IP"
	echo "IP = the IP address or IP name of the working database to recover from"
}

# starting here code:

echo "Database recovery script"

if [ ! -f /var/apresa/conf/opt_cluster ]
then
	echo "This script is only for use within a H.A. cluster"
	norecover
fi

if [ ! -w /recordings ]
then
  echo "This script must run with full permissions"
  norecover
fi

if systemctl is-active pacemaker
then
  echo "Pacemaker is active"

  # check if database of this node is stopped
  psql apresa -c "select count(id) from activecalls;" > /dev/null 2>&1
  if [ "$?" -eq "0" ]
  then
    echo "This node already has an active database"
    norecover
  fi

  crm_mon -A -1 | grep postgresql-status | grep -e PRI -e "HS:alone" > /dev/null 2>&1
  if [ "$?" -ne "0" ]
  then
    echo "There is no other node with a working database"
    norecover
  fi

else
  echo "Pacemaker is not active"
fi


if [ "$1" = "" ]
then
	echo "No IP address specified"
	printusage
	norecover
fi
DBIP=$1

echo "Test if reachable"
ping -c1 $DBIP > /dev/null
if [ "$?" -ne "0" ]
then
  echo "Host $DBIP is not reachable"
  norecover
fi

grep $DBIP /etc/hosts | grep `hostname`
if [ "$?" -eq "0" ]
then
  echo "Specified $DBIP is the current computer"
  norecover
fi

echo "Load database from: $DBIP"
echo "WARNING: This will destroy the existing data in the database in this node"
read -r -p 'Continue? [y/N]' response
case $response in
	[yY])
	:
	;;
	*)
	norecover
	;;
esac

if systemctl is-active pacemaker
then
  echo "Stopping pacemaker"
  service pacemaker stop
  # OLD: it seems this causes problem when an immediate move to this server is done
  # echo "Putting node in standby mode"
  # crm node standby
fi

echo "Stopping database"
service postgresql stop

echo "Clearing local data"
rm -r /database/db/main/*
rm -r /database/db/pg_archive/*

echo "Loading database from $DBIP"
sudo -u postgres pg_basebackup -h $DBIP -U postgres -D /database/db/main -X stream -P
if [ "$?" -ne "0" ]
then
	echo "pg_basebackup failed"
	norecover
fi


echo "Completing"
mkdir /var/run/postgresql
chown postgres:postgres /var/run/postgresql
ln -s /etc/ssl/certs/ssl-cert-snakeoil.pem /database/db/main/server.crt
ln -s /etc/ssl/private/ssl-cert-snakeoil.key /database/db/main/server.key
rm /var/lib/pgsql/tmp/PGSQL.lock

if systemctl is-active pacemaker
then
  # should not happen because stopped above

  # do not restart postgresql ourselves
  # pacemaker agent needs to do it (to create recovery.conf)
  # this is needed to recover from an error
  echo "Reset error status"
  crm resource cleanup postgresql
  crm resource cleanup ms_postgresql
  echo "Bringing the node back online"
  crm node online
else
  echo "INFO: start pacemaker service to bring node online,"
  echo "when this node is ready for it,"
  echo "using the following command:"
  echo "   sudo service pacemaker start"
  read -r -p 'Start pacemaker now? [y/N]' response
  case $response in
    [yY])
    echo "Starting pacemaker now"
    service pacemaker start
    crm configure delete loc_`hostname`
    ;;
  esac
fi

echo "Recovery script finished"

