<?php
/*
Transcription service (Watson, Voicecrunch, etc.)
Processing of calls after completion.

Configuration part 'scribe', prefix 'Scribe_'.
Because we call this only once per minute,
we need to handle multiple at a time,
so we use a queue.

*/
require_once __DIR__ . "/" . "../conf/voip_config.php";
require_once __DIR__ . "/" . CMD_WEB_RELPATH . "db.php";
require_once __DIR__ . "/" . CMD_WEB_RELPATH . "signaling.php";
require_once __DIR__ . "/" . CMD_WEB_RELPATH . "utility.php";
require_once __DIR__ . "/" . CMD_WEB_RELPATH . "datafile.php";
require_once __DIR__ . "/" . CMD_WEB_RELPATH . "encryption.php";

define('QSTATUS_NORMAL',0);
// unused: define('QSTATUS_BUSY',1);
define('QSTATUS_MAXDUR',2);
// unused: define('QSTATUS_DONE',3); // we do not really know

define('TSTATUS_INIT',0);
define('TSTATUS_BUSY',1);
define('TSTATUS_DONE',2);
define('TSTATUS_ERROR',3);

define('CALLSTATUS_PERMFAIL',1);
define('CALLSTATUS_TEMPFAIL',2);
define('CALLSTATUS_NOTREADY',3);
define('CALLSTATUS_DONE',4);

abstract class CrunchStatus {
  const Queued = 0;
  const InProgress = 1;
  const Processed = 2;
  const Failed = 3;
  const QuotaLimit = 4;
  const NotFound = 5;
  const KeywordMatch = 6;
}

// maximum items in the queue
define('MAXQLEN',300);
// maximum duration of calls in the queue in seconds
define('MAXQDUR',24*60*60);

function vlg($t)
{
  // verbose logging
  if (defined('VERBOSE_LOGING'))
  {
    lg($t);
  }
  return;
}

class Transcriber
{
  // Properties
  var $q;
  var $qsize;
  var $durq;
  var $cfg;
  
  
  function GetSetting($key)
  {
    return str_ensure_defined( $this->cfg['Scribe_' . $key] );
  }
  
  function GetIntSetting($key)
  {
    return int_ensure_defined( $this->cfg['Scribe_' . $key] );
  }
  
  function SetSetting($key,$value,$encrypted)
  {
    global $dbintf;
    $dbValue = $value;
    if ($encrypted) $dbValue = encrypt($dbValue);
    $dbintf->SetConfigsPart(array('Scribe_' . $key => $dbValue),'scribe');
    $this->cfg['Scribe_' . $key] = $value;
  }
  
  function Crunch_ExecWithRetry(&$ref_cu)
  {
    $response = curl_exec($ref_cu);
    if ($response === false) return $response;
    $responseCode = curl_getinfo($ref_cu, CURLINFO_HTTP_CODE);
    if ($responseCode != 200)
    {
      lg('crunch response code ' . $responseCode . ' , response: ' . $response);
    }
    if ($responseCode == 401)
    {
      // needs login
      lg('crunch Try get token');
      $Server = $this->GetSetting('Server');
      $Usr = $this->GetSetting('Usr');
      $Pwd = $this->GetSetting('Pwd');
      // do not log passwords: lg($Usr . '/' . $Pwd);
      $cu = curl_init();
      curl_setopt($cu,CURLOPT_URL,$Server . "/api/v1/auth/token");
			curl_setopt($cu,CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
      curl_setopt($cu,CURLOPT_RETURNTRANSFER, true);
      curl_setopt($cu,CURLOPT_POST, 1);
      curl_setopt($cu,CURLOPT_HTTPHEADER, array
        ( 'Accept: application/json'
        , 'Content-Type: application/json'
        ) );
      curl_setopt($cu,CURLOPT_POSTFIELDS,json_encode(array
        ( 'username' => $Usr
        , 'password' => $Pwd
        )) );
      $authResponse = curl_exec($cu);
      if ($authResponse === false)
      {
        lg('crunch auth token request failed');
        return false;
      }
      $authResponseCode = curl_getinfo($cu, CURLINFO_HTTP_CODE);
      curl_close($cu);
      if ($authResponseCode != 200)
      {
        lg('crunch response code ' . $authResponseCode . ' , response: ' . $authResponse);
        return false;
      }
      $authObj = json_decode( $authResponse );
      if ($authObj==null)
      {
        lg('crunch token reply not valid json: ' . $authResponse);
        return false;
      }
      if (isset($authObj->token)==false)
      {
        lg('crunch token reply lacks token');
        return false;
      } 
      $ApiKey = $authObj->token;
      lg('crunch New token of size ' . strlen($ApiKey) );
      vlg('crunch New token: ' . $ApiKey );
      $this->SetSetting('ApiKey',$ApiKey,true);
      // retry with new token
      curl_setopt($ref_cu,CURLOPT_HTTPHEADER, array
      ( 'Authorization: Bearer ' . $ApiKey
      , 'Accept: application/json'
      , 'Content-Type: application/json'
      ) );
      //curl_setopt($ref_cu, CURLOPT_HEADER, 1);
      $response = curl_exec($ref_cu);
      if ($response === false) return $response;
      $responseCode = curl_getinfo($ref_cu, CURLINFO_HTTP_CODE);
      if ($responseCode != 200)
      {
        lg('crunch response code ' . $responseCode . ' , response: ' . $response);
      }
    }
    return $response;
  }
  
  function Crunch_StartingHandlingCall($thecall)
  {
		global $dbintf;
		
    $ret = array();
    $ret['status'] = NULL;
    
    $fullfn = AUDIO_PATH . '/' . $thecall['directory'] . '/' . $thecall['filename'];
    $ApiKey = $this->GetSetting('ApiKey');
    $Server = $this->GetSetting('Server');
    vlg('Server: ' . $Server);
    if ($Server=='')
    {
      vlg('tempfail: server config empty');
      $ret['errmsg'] = 'server config empty';
      $ret['status'] = CALLSTATUS_TEMPFAIL;
      return $ret;
    }

    vlg('crunch upload ' . $fullfn);
		
		$callinfo = $dbintf->GetCallInfo($thecall['id']);
    
    $dataBase64 = base64_encode( file_get_contents($fullfn) );
    
    $cu = curl_init();
    curl_setopt($cu,CURLOPT_URL,$Server . "/api/v1/speech/process");
		curl_setopt($cu,CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    curl_setopt($cu,CURLOPT_RETURNTRANSFER, true);
    curl_setopt($cu,CURLOPT_POST, 1);
    curl_setopt($cu,CURLOPT_HTTPHEADER, array
      ( 'Authorization: Bearer ' . $ApiKey
      , 'Accept: application/json'
      , 'Content-Type: application/json'
      ) );
    $body = json_encode(array
      ( 'dataBase64' => $dataBase64
      , 'filename' => $thecall['filename']
			, 'metadata' => json_encode(array
				( 'local' => $callinfo['local']
				, 'remote' => $callinfo['remote']
				))
      ));
    curl_setopt($cu,CURLOPT_POSTFIELDS,$body);
    $response = $this->Crunch_ExecWithRetry($cu);
    if ($response === false)
    {
      $curl_errmsg = curl_error($cu);
      lg('crunch upload tempfail: could not perform request: ' . $curl_errmsg);
      $ret['errmsg'] = 'could not perform request: ' . $curl_errmsg;
      $ret['status'] = CALLSTATUS_TEMPFAIL;
      curl_close($cu);
      return $ret;
    }
    $retobj = json_decode( $response );
    curl_close($cu);
    
    if ($retobj==null)
    {
      lg('crunch upload tempfail: no valid json reply from server, response was: ' . $response);
      $ret['errmsg'] = 'no valid reply from server';
      $ret['status'] = CALLSTATUS_TEMPFAIL;
      return $ret;
    }
    if ( isset($retobj->errors) )
    {
      $logt = 'crunch upload permfail: errors.';
      if (isset($retobj->message))
      {
        $logt .= ' ' . $retobj->message;
        $ret['errmsg'] = 'upload error: ' . $retobj->message;
      } else
      {
        $ret['errmsg'] = 'error in upload';
      }
      lg( $logt );
      $ret['status'] = CALLSTATUS_PERMFAIL;
      return $ret;
    }
    if ( isset($retobj->id)==false )
    {
      $ret['status'] = CALLSTATUS_PERMFAIL;
      if (isset($retobj->message))
      {
        lg('crunch upload permfail: msg instead of info: ' . $retobj->message);
        $ret['errmsg'] = 'upload error: ' . $retobj->message;
        if ($retobj->message=='Unauthenticated.')
        {
          // retry - user should change config
          $ret['status'] = CALLSTATUS_TEMPFAIL;
        }
      } else
      {
        lg('crunch upload permfail: missing info in reply');
        $ret['errmsg'] = 'error in upload: missing info in reply';
      }
      return $ret;
    }
    
    $ret['status'] = CALLSTATUS_DONE;
    $ret['callid'] = $retobj->id;
    
    return $ret;
  }
  
  function StartHandlingCall(&$plan,$thecall)
  {
    global $dbintf;

    // check if maximum q size reached
    if ($this->qsize + 1 > MAXQLEN)
    {
      // this is not an error
      vlg('q count max reached');
      return false;
    }

    $callinfo = $dbintf->GetCallInfo($thecall['id']);
    $dur = $this->GetDuration($callinfo);
    
    // check if global q duration maximum reached
    
    $globalDurNext = $this->durq + $dur;
    if ($globalDurNext > MAXQDUR && $this->qsize!=0)
    {
      // this is not an error
      vlg('q duration max reached');
      return false;
    }

    // check if plan max duration is reached
    
    $durnext = $plan['durqueued'] + $dur;
    if ($plan['maxdur']!=-1)
    {
      if ($durnext > $plan['maxdur'])
      {
        if ($plan['qstatus'] != QSTATUS_MAXDUR)
        {
          lg('Duration of next call would go past maximum of plan (' . $plan['maxdur'] . ' s), stopping');
          $dbintf->ModifyTScriptPlan($plan['id'],array('qstatus' => QSTATUS_MAXDUR));
        }
        return false;
      }
    }
    if ($plan['qstatus'] == QSTATUS_MAXDUR)
    {
      $dbintf->ModifyTScriptPlan($plan['id'],array('qstatus' => QSTATUS_NORMAL));
    }
    
    lg('Start handling call ' . $callinfo['remoteid'] . ' (ID ' . $thecall['id'] . ')' . ' (' . $dur . ' s)' );
    
    $idext = '';
    
    if ($this->GetSetting('Type')=='crunch')
    {
      $info = $this->Crunch_StartingHandlingCall($thecall);
      if (isset($info['errmsg']))
      {
        vlg('crunch start handling call error: ' . $info['errmsg'] );
        if ($plan['tstatus'] != TSTATUS_ERROR
        || $plan['errmsg'] != $info['errmsg'])
        {
          $plan['tstatus'] = TSTATUS_ERROR;
          $plan['errmsg'] = $info['errmsg'];
          $dbintf->ModifyTScriptPlan($plan['id']
            , array( 'errmsg' => $plan['errmsg']
              , 'tstatus' => $plan['tstatus']
              ) 
            );
        }
      }
      if ($info['status']==CALLSTATUS_TEMPFAIL) return false;
      else if ($info['status']==CALLSTATUS_PERMFAIL)
      {
        // skip this call
      } else if ($info['status']==CALLSTATUS_DONE)
      {
        // upload ok
        $idext = $info['callid'];
        lg('crunch upload ok, extid = ' . $idext);
      }
    }
    
    if ($this->GetSetting('Type')=='sim')
    {
      $idext = substr(md5(rand()), 0, 7);

      if (defined('DEBUG_ERRSIM'))
      {
        // insert an error for testing
        $info = array( 'errmsg' => 'abc problem with ' . $idext );
        $idext = '';
        if ($plan['tstatus'] != TSTATUS_ERROR
        || $plan['errmsg'] != $info['errmsg'])
        {
          $plan['tstatus'] = TSTATUS_ERROR;
          $plan['errmsg'] = $info['errmsg'];
          $dbintf->ModifyTScriptPlan($plan['id']
            , array( 'errmsg' => $plan['errmsg']
              , 'tstatus' => $plan['tstatus']
              ) 
            );
        }
      }
      
    }

    $newdata = array
    ( 'idqueued' => $thecall['id']
    );    
    
    // empty idext means permanent failure and skip
    if ($idext=='')
    {
      $dbintf->ModifyTScriptPlan($plan['id'],$newdata);
      return NULL; // continue with others
    }

    $data = array();
    $data['planid'] = $plan['id'];
    $data['callid'] = $thecall['id'];
    $data['duration'] = $dur;
    $data['idext'] = $idext;
    $data['added'] = time();
    $dbintf->AddTScriptQ($data);
    $this->qsize ++;
    
    $plan['durqueued'] = $durnext;
    if ($plan['tstatus']==TSTATUS_INIT || $plan['tstatus']==TSTATUS_DONE)
    {
      $newdata['tstatus'] = TSTATUS_BUSY;
    }
    $newdata['durqueued'] = $plan['durqueued'];
    $dbintf->ModifyTScriptPlan($plan['id'],$newdata);
    return true;
  }
  
  function Crunch_IsReady($IdxBusy)
  {
    $ret = array();
    // status: see CALLSTATUS_ constants
    $ret['status'] = NULL;
    
    $ApiKey = $this->GetSetting('ApiKey');
    $Server = $this->GetSetting('Server');
    
    $cu = curl_init();
    curl_setopt($cu,CURLOPT_URL,$Server . "/api/v1/calls/" . $IdxBusy . "/status" );
		curl_setopt($cu,CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    curl_setopt($cu,CURLOPT_RETURNTRANSFER, true);
    curl_setopt($cu,CURLOPT_HTTPHEADER, array
      ( 'Authorization: Bearer ' . $ApiKey
      , 'Accept: application/json'
      , 'Content-Type: application/json'
      ) );
    $response = $this->Crunch_ExecWithRetry($cu);
    if ($response === false)
    {
      $curl_errmsg = curl_error($cu);
      lg('crunch upload tempfail: could not perform request: ' . $curl_errmsg);
      $ret['errmsg'] = 'could not perform request: ' . $curl_errmsg;
      $ret['status'] = CALLSTATUS_TEMPFAIL;
      curl_close($cu);
      return $ret;
    }

    $retobj = json_decode( $response );
    curl_close($cu);
    
    if ($retobj==null)
    {
      lg('crunch upload tempfail: no valid json reply from server, response was: ' . $response);
      $ret['errmsg'] = 'no valid reply from server';
      $ret['status'] = CALLSTATUS_TEMPFAIL;
      return $ret;
    }
    
    if (isset($retobj->status)==false)
    {
      lg('crunch callcheck permfail: call status not found');
      $ret['status'] = CALLSTATUS_PERMFAIL;
      return $ret;
    }
    
    if ($retobj->status==CrunchStatus::Failed
    || $retobj->status==CrunchStatus::NotFound
    ) // error
    {
      lg('crunch callcheck permfail: call has error status ' . $retobj->status . ' at server: ' . $retobj->statusDescription);
      $ret['status'] = CALLSTATUS_PERMFAIL;
      return $ret;
    }
    
    // we have a valid reply
    if ($retobj->status != CrunchStatus::Processed)
    {
      vlg("crunch callcheck not ready, call status: " . $retobj->status );
      $ret['status'] = CALLSTATUS_NOTREADY;
      return $ret;
    }

    $ret['status'] = CALLSTATUS_DONE;
    
    $ret = array_merge($ret,$this->LoadTranscription($retobj));
    
    return $ret;
  }
  
  function filled($channel)
  {
    if (is_array($channel)==false) return false;
    if (count($channel)==0) return false;
    return true;
  }

  function LoadTranscription($retobj)
  {
    $ret = [];
    // no transcript?
    if (isset($retobj->transcription)==false) return $ret;
    $LR = $this->filled($retobj->transcription->leftChannel)
    && $this->filled($retobj->transcription->rightChannel);
    $Both = $this->filled($retobj->transcription->bothChannels);
    if ($LR==false && $Both==false) return $ret;
    
    $streams = [];
    $mainLanguage = $retobj->transcription->language;
    $streams[$mainLanguage] = new TranscriptStream();
    $translateLng = '';
    
    if ($LR)
    {
      $L = $retobj->transcription->leftChannel;
      $R = $retobj->transcription->rightChannel;
      $Lix = $Rix = 0;
      $Lcnt = count($L);
      $Rcnt = count($R);
      while ($Lix<$Lcnt || $Rix<$Rcnt)
      {
        $toadd = '';
        if ($Lix<$Lcnt && ($Rix>=$Rcnt || $L[$Lix]->start <= $R[$Rix]->start) )
        {
          // use L
          $lng = $L[$Lix]->language;
          $newCh = 'A';
          $toadd = $L[$Lix]->content;
          $Lix ++;
        } else
        {
          // use R
          $lng = $R[$Rix]->language;
          $newCh = 'B';
          $toadd = $R[$Rix]->content;
          $Rix ++;
        }
        if (isset($streams[$lng])==false)
        {
          $streams[$lng] = new TranscriptStream();
          if ($translateLng=='') $translateLng = $lng;
        }
        echo $lng . ' ' . $toadd . "\n";
        $streams[$lng]->add($toadd, $newCh);
      }
    } else if ($Both)
    {
      $ts = $retobj->transcription->bothChannels;
      foreach ($ts as $wordobj)
      {
        if (isset($wordobj->content)==false) continue;
        $lng = $wordobj->language;
        $content = $wordobj->content;
        if ($content=='') continue;
        if (isset($streams[$lng])==false)
        {
          $streams[$lng] = new TranscriptStream();
          if ($translateLng=='') $translateLng = $lng;
        }
        $streams[$lng]->add($content, false);
      }
    }
    $ret['transcript'] = $streams[$mainLanguage]->get();
    if ($translateLng!='') $ret['translation'] = $streams[$translateLng]->get();
    
    return $ret;
  }
  
  function SetCallData($id,$target,$data)
  {
    global $dbintf;
    if ($target=='transcript')
    {
      $dbintf->SetTranscript($id,$data);
    } else if ($target=='notes'
    || $target=='notes2'
    || $target=='notes3'
    || $target=='notes4'
    || $target=='notes5'
    || $target=='data1'
    || $target=='data2'
    || $target=='data3'
    || $target=='data4'
    || $target=='data5'
    )
    {
      $dbintf->SetCallInfo($id,array($target=>$data));
    }    
  }
  
  function IsReady($IdBusy,$IdxBusy)
  {
    global $dbintf;
    $ret = false;
    if ($this->GetSetting('Type')=='crunch')
    {
      $info = $this->Crunch_IsReady($IdxBusy);
      if ($info['status']==CALLSTATUS_NOTREADY) return false;
      if ($info['status']==CALLSTATUS_TEMPFAIL) return false;
      // do not check again if permanent error
      if ($info['status']==CALLSTATUS_PERMFAIL) return NULL;
      if ($info['status']==CALLSTATUS_DONE)
      {
        $ret = true;
        if (isset($info['transcript']))
        {
          // store transcript somewhere
          $target = $this->GetSetting('Target');
          vlg('target = ' . $target);
          lg('tplan transcript of length ' . strlen($info['transcript']) );
          if ($target=='') $target = 'transcript';
          $this->SetCallData($IdBusy,$target,$info['transcript']);
        } else lg('tplan no transcript');
        if (isset($info['translation']))
        {
          // store translation somewhere?
          $target = $this->GetSetting('TranslationTarget');
          vlg('translation target = ' . $target);
          lg('tplan translation of length ' . strlen($info['translation']) );
          $this->SetCallData($IdBusy,$target,$info['translation']);
        }
      } else return NULL; // other status should never happen
    }
    if ($this->GetSetting('Type')=='sim')
    {
      $r = rand(1,5);
      //vlg('R = ' . $r);
      if ($r==1) $ret = true;
      else $ret = false;
    }
    return $ret;
  }

  function dumpCfg()
  {
    if (defined('DEBUG_LOGING'))
    {
      foreach ($this->cfg as $key => $value)
      {
        $key = substr($key,7);
        if (strlen($value)>50) $value = substr($value,0,47) . '...';
        vlg('- ' . $key . ' = ' . $value);
      }
    }
  }

  // duration in seconds
  function GetDuration($callinfo)
  {
    $dur = $callinfo['duration'];
    if (isset($callinfo['durms']))
    {
      $dur += $callinfo['durms'] / 1000.0;
    }
    return $dur;
  }
  
  function PlanQStatusStr($qstatus)
  {
    if ($qstatus==0) return 'normal';
    // unused: else if ($qstatus==QSTATUS_BUSY) return 'in progress';
    else if ($qstatus==QSTATUS_MAXDUR) return 'max duration';
    // unused: else if ($qstatus==QSTATUS_DONE) return 'completed';
    return '' . $qstatus;
  }

  function PlanTStatusStr($tstatus)
  {
    if ($tstatus==TSTATUS_INIT) return 'init';
    else if ($tstatus==TSTATUS_BUSY) return 'in progress';
    else if ($tstatus==TSTATUS_DONE) return 'completed';
    return '' . $tstatus;
  }
  
  function GetItemsInQForPlan($planid)
  {
    $ret = array();
    foreach ($this->q as $item)
    {
      if ($item['planid'] == $planid)
      {
        $ret[] = $item;
      }
    }
    return $ret;
  }
  
  function CheckReadyForPlan($plan)
  {
    global $dbintf;
    
    $noneLeftInQ = true;
    $qItems = $this->GetItemsInQForPlan($plan['id']);
    foreach ($qItems as $item)
    {
      $IdBusy    = $item['callid'];
      $IdxBusy   = $item['idext'];
      
      // something is brewing (busy)
      $result = $this->IsReady($IdBusy,$IdxBusy);
      if ($result===true || $result===NULL)
      {
        // ready or permanent failure
        $dur = $item['duration'];
        $newdata = array();
        if ($result===true) // ready
        {
          $plan['durdone'] += $dur;
          $plan['cdone'] ++;
          $newdata['durdone'] = $plan['durdone'];
          $newdata['cdone'] = $plan['cdone'];
          $newdata['tstatus'] = TSTATUS_BUSY;
          lg("Processing ready of call $IdBusy ($dur s). Total this plan: " . $plan['durdone'] . ' s');
        } else // NULL
        {
          $newdata['tstatus'] = TSTATUS_ERROR;
          lg("Permanent failure skipping call $IdBusy");
        }
        $IdHandled = $IdBusy;
        $dbintf->DelTScriptQ($item['id']);
        $this->qsize --;
        $this->durq -= $dur;
        vlg('UPD. Queue: ' . $this->durq . ' s in ' . $this->qsize . ' items');
        $dbintf->ModifyTScriptPlan($plan['id'],$newdata);
      } else
      {
        vlg('Still busy with ' . $IdBusy);
        // because we can have multiple in q, continue
        $noneLeftInQ = false;
      }
    }
    
    if ($noneLeftInQ)
    {
      // we are done
      if ($plan['tstatus'] == TSTATUS_BUSY)
      {
        // and we were busy
        vlg('transcription for this plan is done for now');
        $newdata = array();
        $newdata['tstatus'] = TSTATUS_DONE;
        $dbintf->ModifyTScriptPlan($plan['id'],$newdata);
      }
    } else
    {
      if ($plan['tstatus'] != TSTATUS_BUSY)
      {
        vlg('transcription for this plan is busy');
        $newdata = array();
        $newdata['tstatus'] = TSTATUS_BUSY;
        $dbintf->ModifyTScriptPlan($plan['id'],$newdata);
      }
    }    
  }
  
  function CheckAddForPlan($plan)
  {
    global $dbintf;
    
    $IdQueued = $plan['idqueued'];
    // vlg('ID queued: ' . $IdQueued );
    
    $latest = $dbintf->GetIdOfLatestCall();
    if ($latest > $IdQueued)
    {
      // we have new recordings
      $loopix = 0;
      while ($latest > $IdQueued)
      {
        $loopix++;
        // safety break
        if ($loopix > MAXQLEN*2) break;
        
        $limit = 1;
        $cfg = array();
        $filter = StringToArray($plan['filter']);
        foreach ($filter as $key=>$value)
        {
          if ($key=='fd') $cfg['from_ts'] = $value;
          if ($key=='ld')
          {
            $ld = (int) $value;
            // next day
            $ld = $ld + 60 * 60 * 24;
            $cfg['to_ts'] = $ld;
          }
          if ($key=='id') $cfg['telnum'] = $value;
          if ($key=='fdur') $cfg['fdur'] = $value;
          if ($key=='ldur') $cfg['ldur'] = $value;
        }
        $calls = $dbintf->GetToHandleList($IdQueued,$limit,$cfg);
        if (count($calls)==0)
        {
          // there could be still in q
          vlg('tplan: no more matching calls');
          $IdQueued = $latest;
          $newdata = array();
          $newdata['idqueued'] = $IdQueued;
          $dbintf->ModifyTScriptPlan($plan['id'],$newdata);
          break;
        } else
        {
          $thecall = $calls[0];
          $IdQueued = $thecall['id'];
          $result = $this->StartHandlingCall( $plan, $thecall );
          if ($result===NULL || $result===false)
          {
            // false is temp failure where the same call is retried
            // NULL is permanent failure where call is skipped
            // in both cases do not try next one immediately
            // something went wrong or limit reached or other problem
            break;
          }
        }
      } // while loop
    } else
    {
      // no new calls
      // but new ones could be added later
    }
  }
  
  function PerformPlan($plan)
  {
    global $dbintf;
    
    // disabled plans are skipped
    if ($plan['enabled']==0) return;
    
    $name = $plan['name'];
    $qstatus = (int) $plan['qstatus'];
    $tstatus = (int) $plan['tstatus'];
    vlg("Plan: " . $plan['name'] . ' (' . (double) $plan['durdone'] . ' s)'
    . ' qstatus: ' . $this->PlanQStatusStr($qstatus) 
    . ' , tstatus: ' . $this->PlanTStatusStr($tstatus) 
    );
    
    $this->CheckReadyForPlan($plan);
    $this->CheckAddForPlan($plan);

  }

  function LoadCfg()
  {
    global $dbintf;
    $this->cfg = $dbintf->LoadConfigsPart('scribe');
    if (isset($this->cfg['Scribe_Pwd']))
    {
      $this->cfg['Scribe_Pwd'] = decrypt( $this->cfg['Scribe_Pwd'] );
    }
    if (isset($this->cfg['Scribe_ApiKey']))
    {
      $this->cfg['Scribe_ApiKey'] = decrypt( $this->cfg['Scribe_ApiKey'] );
    }
    if (defined('VERBOSE_LOGING'))
    {
      $this->dumpCfg();
    }
  }
  
  function LoadQ()
  {
    global $dbintf;
    // load queue
    $this->q = $dbintf->GetTScriptQ();
    
    // total items in the queue
    $this->qsize = count( $this->q );
    // total duration of items in the queue
    $this->durq = 0.0;
    foreach ($this->q as $item)
    {
      $this->durq += $item['duration'];
    }
    vlg('Queue: ' . $this->durq . ' s in ' . $this->qsize . ' items');
  }
    
  function Run()
  {
    global $dbintf;
    $this->LoadCfg();
    
    $handleType = $this->GetSetting('Type');
    vlg('tplan handle type: ' . $handleType);
    if ($handleType=='') return;
    if ($handleType=='sim'
    ||  $handleType=='crunch') ;
    else {
      // should never happen
      lg('tplan unsupported handle type');
      return;
    }
    
    $this->LoadQ();

    $plans = $dbintf->GetTScriptPlans();
    foreach ($plans as $plan)
    {
      $this->PerformPlan($plan);
    }
    
  }
  
}

class TranscriptStream
{
  var $curCh = false;
  var $text = '';
  function add($newText, $newCh)
  {
    if ($newCh !== false)
    {
      // multi-channel
      if ($this->curCh!=$newCh) $this->text .= ' ' . $newCh  . ':';
      $this->curCh = $newCh;
    }
    if ($this->text!='' && ctype_alnum($newText[0])) $this->text .= ' ';
    $this->text .= $newText;
  }
  function get()
  {
    if (strlen($this->text)>1 && $this->text[0]==' ') $this->text = substr($this->text,1);
    return $this->text;
  }
}
