var call_encryption_available = crypto != undefined && crypto.subtle != undefined;

function ab2str(buf) 
{
  return String.fromCharCode.apply(null, new Uint8Array(buf));
}

function str2ab(str)
{
  return new Uint8Array(str.split("").map( (c) => c.charCodeAt())).buffer;
}

async function DerivePrivKeyWrapKey(password, salt)
{
  let enc = new TextEncoder();

  let keyMaterial = await crypto.subtle.importKey(
    "raw",
    enc.encode(password),
    "PBKDF2",
    false,
    ["deriveKey", "deriveBits"]
  );

  return crypto.subtle.deriveKey(
    {
      name: "PBKDF2",
      salt: salt,
      iterations: 100000,
      hash: "SHA-256"
    },
    keyMaterial,
    { "name": "AES-GCM", "length": 256},
    false,
    ["wrapKey", "unwrapKey"]
  );
}

async function WrapPrivKey(privKey, wrapKey, iv)
{
  return await crypto.subtle.wrapKey(
    "pkcs8",
    privKey,
    wrapKey,
    {
      name: "AES-GCM",
      iv: iv
    }
  );
}

async function UnwrapPrivKey(privKey, wrapKey, iv, exportable)
{
  try
  {
    return await crypto.subtle.unwrapKey(
      "pkcs8",
      privKey,
      wrapKey,
      {
        name: "AES-GCM",
        iv: iv
      },
      {
        name: "RSA-OAEP",
        hash: "SHA-1"
      },
      exportable,
      ["decrypt"]
    );
  }
  catch(ex)
  {
    // This one throws an empty exception when the password was not correct...
    throw new DOMException("<%~Web.WrongPassword%>");
  }
}

async function CreateNewCallEncryptionKey(password)
{
  let keypair = await crypto.subtle.generateKey(
    {
      name: "RSA-OAEP",
      modulusLength: 4096,
      publicExponent: new Uint8Array([0x01, 0x00, 0x01]),
      hash: "SHA-1"
    },
    true,
    ["encrypt", "decrypt"]
  );

  let salt = crypto.getRandomValues(new Uint8Array(16));
  let iv = crypto.getRandomValues(new Uint8Array(12));
  let wrapKey = await DerivePrivKeyWrapKey(password, salt);
  let privKeyWrapped = await WrapPrivKey(keypair.privateKey, wrapKey, iv);

  let pubKey = await crypto.subtle.exportKey(
    "spki",
    keypair.publicKey
  );

  let pubKeyPem = "-----BEGIN PUBLIC KEY-----\n" + btoa(ab2str(pubKey)) + "\n-----END PUBLIC KEY-----";

  let out = {
    privKey: btoa(ab2str(privKeyWrapped)),
    pubKey: pubKeyPem,
    salt: btoa(ab2str(salt)),
    iv: btoa(ab2str(iv)),
  }

  return out;
}

async function FetchPrivKey()
{
  params = new URLSearchParams;
  params.append('data', 'privkey');
  params.append('SessionToken', SessionToken);

  let response = await fetch('/getdata.php',
  {
    method: 'POST',
    body: params
  });

  if(response.ok)
  {
    return await response.json();
  }
  else
  {
    throw new DOMException("Could not fetch the key: " + response.statusText, "NetworkError");
  }
}

async function GetPrivKey(password)
{
  let key = await FetchPrivKey();
  let wrapKey = await DerivePrivKeyWrapKey(password, str2ab(atob(key.salt)));
  let privKey = await UnwrapPrivKey(str2ab(atob(key.privKey)), wrapKey, str2ab(atob(key.iv)), true);
  return privKey;
}

async function WrapKeyNewPassword(privKey, password)
{
  let salt = crypto.getRandomValues(new Uint8Array(16));
  let iv = crypto.getRandomValues(new Uint8Array(12));
  let wrapKey = await DerivePrivKeyWrapKey(password, salt);
  let privKeyWrapped = await WrapPrivKey(privKey, wrapKey, iv);
  let out = {
    privKey: btoa(ab2str(privKeyWrapped)),
    salt: btoa(ab2str(salt)),
    iv: btoa(ab2str(iv)),
  }
  return out;
}

async function FetchWaveSurfer(url, id)
{
  try
  {
    let call = await FetchCall(url, id);
    mediaPlayer.loadArrayBuffer(call);
  }
  catch(ex)
  {
    if(ex.name != "AbortError")
    {
      if(ex.message.length > 0)
      {
        alert(ex.message);
      }
      else
      {
        alert("<%Web.CallEncryptGenericErr%>");
      }
    }
    prevFile="";
    setPlayImg();
    InitWaveSurferPlayer();
  }
}

async function DownloadEncryptedReal(url, id, filename)
{
  let databuf = await FetchCall(url, id);
  let a = document.createElement("a");
  document.body.appendChild(a);
  a.style = "display: none";
  
  let blob = new Blob([databuf]);
  url = window.URL.createObjectURL(blob);
  a.href = url;
  a.download = filename;
  a.click();
  window.URL.revokeObjectURL(url);
}

async function DownloadEncrypted(url, id, filename)
{
  try
  {
    DownloadEncryptedReal(url, id, filename);
  }
  catch(ex)
  {
    if(ex.name != "AbortError")
    {
      if(ex.message.length > 0)
      {
        alert(ex.message);
      }
      else
      {
        alert("<%Web.CallEncryptGenericErr%>");
      }
    }
  }
}

async function DownloadMultipleEncrypted(rows)
{
  try
  {
    for(let ix = 0; ix < rows.length; ++ix)
    { 
      var id = rows[ix].getData("id");
      var fn = rows[ix].getData("filename");
      await DownloadEncryptedReal('getcall/' + fn + '?id=' + id + "&d=1", id, fn);
    }
  }
  catch(ex)
  {
    if(ex.name != "AbortError")
    {
      if(ex.message.length > 0)
      {
        alert(ex.message);
      }
      else
      {
        alert("<%Web.CallEncryptGenericErr%>");
      }
    }
  }
}

async function FetchKey(password, id)
{
  params = new URLSearchParams;
  params.append('data', 'callkey');
  params.append('SessionToken', SessionToken);
  params.append('callid', id);

  let response = await fetch('/getdata.php',
  {
    method: 'POST',
    body: params
  });

  if(response.ok)
  {
    let keyParts = await response.json();
    let privateKeyKey = await DerivePrivKeyWrapKey(password, str2ab(atob(keyParts["salt"])));
    
    try
    {
      var privKey = await crypto.subtle.unwrapKey(
        "pkcs8",
        str2ab(atob(keyParts["PrivateKey"])),
        privateKeyKey,
        {
          name: "AES-GCM",
          iv: str2ab(atob(keyParts["iv"]))
        },
        {
          name: "RSA-OAEP",
          hash: "SHA-1"
        },
        false,
        ["decrypt"]
      );
    }
    catch(ex)
    {
      // This one throws an empty exception when the password was not correct...
      throw new DOMException("<%~Web.WrongPassword%>");
    }

    let EncCallKey = str2ab(atob(keyParts["CallKey"]));

    let CallKey = await crypto.subtle.decrypt(
      { name: "RSA-OAEP" },
      privKey,
      EncCallKey
    );
    
    return {
      key: CallKey,
      counter: str2ab(atob(keyParts["counter"]))
    };
  }
  else
  {
    throw new DOMException("Failed to retrieve the call key: " + response.statusText, "NetworkError");
  }
}

async function DecryptCallsReal(ids)
{
  try
  {
    let password = await GetCallEncryptionPwd();

    let toDecrypt = [];

    for(let ix = 0; ix < ids.length; ix++)
    {
      let callKey = await FetchKey(password, ids[ix]);
      let info = {
        id: ids[ix],
        key: btoa(ab2str(callKey['key']))
      }

      toDecrypt.push(info);
    }

    let records = JSON.stringify(toDecrypt);

    addData('act', 'calldecrypt');
    addData('records', records);
    ActionSubmit();
  }
  catch(ex)
  {
    if(ex.message.length > 0)
    {
      alert(ex.message);
    }
    else
    {
      alert("<%Web.CallEncryptGenericErr%>");
    }
  }
}

async function FetchCall(url, id)
{
  Dom.get('progress-text-disp').innerText = "0%";
  progressBar.set('value', 0);

  let password = await GetCallEncryptionPwd();
  let callKey = await FetchKey(password, id);

  Dom.get('calldownloadprogress').style = "";
  let response = await fetch(url);
  if(response.ok)
  { 
    const reader = response.body.getReader();
    const contentLength = +response.headers.get('Content-Length');
    let chunksAll = new Uint8Array(contentLength);

    let received = 0;

    while(true)
    {
      const {done, value} = await reader.read();

      if(done) break;
      
      chunksAll.set(value, received)
      received += value.length;

      let progress = Math.round((received / contentLength) * 100);
      Dom.get('progress-text-disp').innerText = progress.toString() + "%";
      progressBar.set("value", progress);
    }
    
    let key = await crypto.subtle.importKey(
      'raw',
      callKey['key'],
      'AES-CTR',
      true,
      ['encrypt', 'decrypt']
    );

    let decrypted = await crypto.subtle.decrypt(
      {
        name: 'AES-CTR',
        counter: callKey['counter'],
        length: 64
      },
      key,
      chunksAll
    );
    
    return decrypted;
  }
  else
  {
    throw new DOMException(response.statusText, "NetworkError");
  }
}

async function SetNewCallEncryptionKey(password)
{
  try
  {
    let newKey = await CreateNewCallEncryptionKey(password);
    addData('act', 'newKey');
    addData('privKey', newKey.privKey);
    addData('salt', newKey.salt);
    addData('iv', newKey.iv);
    addData('pubKey', newKey.pubKey);
    ActionSubmit();
  }
  catch(ex)
  {
    alert('Failed to generate a new keypair: ' + ex.message);
  }
}

async function ChangePassword(currentpwd, newpwd)
{
  try
  {
    let curKey = await GetPrivKey(currentpwd);
    let newKey = await WrapKeyNewPassword(curKey, newpwd);

    addData('act', 'newPrivKeyPwd');
    addData('privKey', newKey.privKey);
    addData('salt', newKey.salt);
    addData('iv', newKey.iv);
    addData('enable', Dom.get('TnCallEncrypt').checked ? '1' : '0');
    ActionSubmit();
  }
  catch(ex)
  {
    alert("<%Web.CallEncryptPwdChangeFail%>" + ex.message);
  }
}