<?php
require_once 'certificates_general.php';
require_once 'db.php';

if(!function_exists('curl_reset'))
{
	// Workaround for curl_reset function being unavailable in Debian 7 due to old PHP
	function curl_reset(&$ch)
	{
		curl_close($ch);
		$ch = curl_init();
	}
}

trait Base64Url
{
  function Base64UrlEncode($input)
  {
    $base64 = base64_encode($input);
    return strtr(rtrim($base64, "="), "+/", "-_");
  }

  function Base64UrlDecode($input)
  {
    $pad = strlen($input) % 4;

    switch($pad)
    {
      case 0:
        break;
      case 2:
        $input .= '==';
        break;
      case 3:
        $input .= '=';
        break;
      default:
        throw new Exception("Illegal base64url string");
    }

    $base64 = strtr($input, "-_", "+/");
    return base64_decode($base64);
  }
};

class HttpException extends RuntimeException
{

};

class HttpTransferException extends HttpException
{
  private $request;

  function __construct($message, $code, $request)
  {
    parent::__construct($message, $code);
    $this->request = $request;
  }

  function GetRequest()
  {
    return $this->request;
  }
};

class HttpBadClientResponse extends HttpTransferException
{
  private $response;

  function __construct($message, $code, $request, $response)
  {
    parent::__construct($message, $code, $request);
    $this->response = $response;
  }

  function GetResponse()
  {
    return $this->response;
  }
};

class HttpBadServerResponse extends HttpTransferException
{
  private $response;

  function __construct($message, $code, $request, $response)
  {
    parent::__construct($message, $code, $request);
    $this->response = $response;
  }

  function GetResponse()
  {
    return $this->response;
  }
};

class HttpOpBase
{
  protected $data;
  function __construct(array $in)
  {
    $this->data = $in;
  }

  function GetHeader($name)
  {
    foreach($this->data["headers"] as $key=>$value)
    {
      if(strcasecmp($key, $name) == 0)
      {
        return $value;
      }
    }
    return null;
  }

  function GetUrl()
  {
    return $this->data["url"];
  }
};

class HttpRequest extends HttpOpBase
{
  function GetMethod()
  {
    return $this->data["method"];
  }
};

class HttpResponse extends HttpOpBase
{
  private $request;

  function __construct(array $response, $request)
  {
    parent::__construct($response);
    $this->request = $request;
  }

  function GetRequest()
  {
    return $this->request;
  }

  function GetStatusCode()
  {
    return $this->data["statusCode"];
  }

  function GetReasonPhrase()
  {
    return $this->data["reasonPhrase"];
  }

  function GetProtocolVersion()
  {
    return $this->data["protocolVersion"];
  }

  function GetBody()
  {
    return $this->data["body"];
  }

  function GetJson()
  {
    if(isset($this->data["body"]))
    {
      $json = json_decode(stream_get_contents($this->data["body"]), true);
      rewind($this->data["body"]);
      return $json;
    }
    return null;
  }
};

class HttpClient
{
  private $ch;
  private $defaultOptions;

  function __construct(array $defaultOptions = [])
  {
    if(!extension_loaded('curl'))
    {
      throw new HttpException("The PHP curl extension is not loaded");
    }
    
    $this->ch = curl_init();
    if($this->ch === false)
    {
      throw new HttpException("Could not allocate curl handle");
    }

    $this->defaultOptions = $defaultOptions;
  }

  private function DefaultCurlOptions(array $request, array &$headers)
  {
    $headersDone = false;

    $options = [
      CURLOPT_CUSTOMREQUEST => $request["method"],
      CURLOPT_URL => $request["url"],
      CURLOPT_RETURNTRANSFER => false,
      CURLOPT_HEADER         => false,
      CURLOPT_CONNECTTIMEOUT => 30,
      CURLOPT_PROTOCOLS =>  CURLPROTO_HTTP | CURLPROTO_HTTPS,
      CURLOPT_HEADERFUNCTION => function($handle, $h) use (&$headersDone, &$headers)
      {
        $header = trim($h);

        if($header == "")
        {
          $headersDone=true;
        }
        else if($headersDone)
        {
          $headers = [$header];
          $headersDone = false;
        }
        else 
        {
          $headers[] = $header;
        }

        return strlen($h);
      },
      CURLOPT_FILE => fopen('php://temp', 'w+')
    ];

    return $options;
  }

  private function MethodCurlOptions(array $request, array &$curlOptions)
  {
    if(isset($request["options"]["body"]))
    {
      $curlOptions[CURLOPT_POSTFIELDS] = $request["options"]["body"];
    }
    else 
    {
      switch($request["method"])
      {
        case 'POST':
        case 'PUT':
        case 'PATCH':
          if(!$this->HasHeader($request, "Content-Length"))
          {
            $curlOptions[CURLOPT_HTTPHEADER][] = "Content-Length: 0";
          }
          break;
        case 'HEAD':
          $curlOptions[CURLOPT_NOBODY] = true;
      }  
    }
  }
  
  private function UserCurlOptions(array $request, array &$curlOptions)
  {
    foreach($request["options"] as $key => $value)
    {
      switch($key)
      {
        case 'verify':
          if(!$value)
          {
            $curlOptions[CURLOPT_SSL_VERIFYHOST] = 0;
            $curlOptions[CURLOPT_SSL_VERIFYPEER] = false;
          }
          break;
        case 'timeout':
          $curlOptions[CURLOPT_TIMEOUT] = $value;
          break;
        case 'connect_timeout':
          $curlOptions[CURLOPT_CONNECTTIMEOUT] = $value;
          break;
        case 'redirect':
          if(is_array($value))
          {
            $curlOptions[CURLOPT_FOLLOWLOCATION] = true;
            $curlOptions[CURLOPT_MAXREDIRS] = isset($value['max']) ? $value['max'] : 5;
            $curlOptions[CURLOPT_POSTREDIR] = isset($value['retain_post']) ? $value['retain_post'] : false;
          }
          else if($value)
          {
            $curlOptions[CURLOPT_FOLLOWLOCATION] = true;
            $curlOptions[CURLOPT_MAXREDIRS] = 5;
          }
          else 
          {
            $curlOptions[CURLOPT_FOLLOWLOCATION] = false;
          }
          break;
        case 'http':
          if(!$value)
          {
            $curlOptions[CURLOPT_PROTOCOLS] &= ~CURLPROTO_HTTP;
          }
          break;
        case 'https':
          if(!$value)
          {
            $curlOptions[CURLOPT_PROTOCOLS] &= ~CURLPROTO_HTTPS;
          }
          break;
      }
    }
  }

  private function AddCurlHeaders(array $request, array &$curlOptions)
  {
    foreach($request['headers'] as $key => $values)
    {
      foreach($values as $value)
      {
        $curlOptions[CURLOPT_HTTPHEADER][] = "$key: $value";
      }
    }
  }

  private function HasHeader(array $request, $header)
  {
    foreach($request['headers'] as $key => $value)
    {
      if(strcasecmp($key, $header) == 0)
      {
        return true;
      }
    }
  }

  private function NormalizeHeaders(array &$request)
  {
    foreach($request['headers'] as $key => $value)
    {
      if(!is_array($value))
      {
        $request['headers'][$key] = [$value];
      }
    }
  }

  private function GetHost($url)
  {
    $host = parse_url($url, PHP_URL_HOST);
    
    if(!$host)
    {
      return false;
    }

    $port = parse_url($url, PHP_URL_PORT);

    if($port)
    {
      $host .= ":$port";
    }

    return $host;
  }

  private function AddDefaultHeaders(&$request)
  {
    if(!$this->HasHeader($request, 'Host'))
    {
      $host = $this->GetHost($request["url"]);

      if($host)
      {
        $request["headers"]["Host"][] = $host;
      }
    }

    if(!$this->HasHeader($request, "User-Agent"))
    {
      $request["headers"]["User-Agent"][] = "Apresa" . ' curl/' . curl_version()["version"] . ' PHP/' . phpversion();
    }
  }

  private function SetupCurl($request)
  {
    curl_reset($this->ch);
    $headers = [];
    $curlOptions = $this->DefaultCurlOptions($request, $headers);
    $this->MethodCurlOptions($request, $curlOptions);
    $this->UserCurlOptions($request, $curlOptions);
    $this->AddCurlHeaders($request, $curlOptions);
    curl_setopt_array($this->ch, $curlOptions);

    return [$curlOptions[CURLOPT_FILE], &$headers];
  }

  private function ThrowTransferException($request)
  {
    $errNo = curl_errno($this->ch);
    $strErr = curl_error($this->ch);

    $message = "Curl error {$errNo}: {$strErr}";
    throw new HttpTransferException($message, curl_errno($this->ch), new HttpRequest($request));
  }

  private function CreateResponse(array $request, $body, array $headers)
  {
    $response = [];

    $versionLine = explode(' ', array_shift($headers), 3);
    
    $response["protocolVersion"] = isset($versionLine[0]) ? substr($versionLine[0], 5) : null;
    $response["statusCode"] = isset($versionLine[1]) ? (int) $versionLine[1] : null;
    $response["reasonPhrase"] = isset($versionLine[2]) ? $versionLine[2] : null;

    if(is_null($response["statusCode"]))
    {
      throw new HttpTransferException("Could not parse HTTP response status-line", 0, new HttpRequest($request));
    }

    $response["headers"] = [];
    $response["url"] = $request["url"];

    foreach ($headers as $header) 
    {
      $parts = explode(':', $header, 2);
      $response["headers"][trim($parts[0])][] = isset($parts[1]) ? trim($parts[1]) : null;
    }

    rewind($body);
    $response["body"] = $body;
    return new HttpResponse($response, new HttpRequest($request));
  }

  function Get($url, array $options = [])
  {
    return $this->Send("GET", $url, $options);
  }

  function Head($url, array $options = [])
  {
    return $this->Send("HEAD", $url, $options);
  }

  function Post($url, array $options = [])
  {
    return $this->Send("POST", $url, $options);
  }

  function Send($method, $url, array $options = [])
  {
    $request = [
      "method" => $method,
      "url" => $url,
      "options" => array_merge($this->defaultOptions, $options)
    ];

    if(isset($request["options"]["headers"]))
    {
      $request["headers"] = $request["options"]["headers"];
    }
    else 
    {
      $request["headers"] = [];
    }

    $this->NormalizeHeaders($request);
    $this->AddDefaultHeaders($request);

    $result = $this->SetupCurl($request);

    $body = $result[0];
    $headers = &$result[1];

    $out = curl_exec($this->ch);

    if($out === false)
    {
      $this->ThrowTransferException($request);
    }

    $response = $this->CreateResponse($request, $body, $headers);

    if($response->GetStatusCode() >= 400)
    {
      $request = new HttpRequest($request);
      $message = "{$request->GetMethod()} on {$request->GetUrl()} returned {$response->GetStatusCode()}";
      
      if($response->GetStatusCode() >= 500)
      {
        throw new HttpBadServerResponse($message, $response->GetStatusCode(), $request, $response);
      }
      else 
      {
        throw new HttpBadClientResponse($message, $response->GetStatusCode(), $request, $response);
      }
    }
    else if($response->GetStatusCode() >= 600)
    {
      $request = new HttpRequest($request);
      $message = "{$request->GetMethod()} on {$request->getUrl()} returned an unknown HTTP status code {$response->GetStatusCode()}";
      throw new HttpBadServerResponse($message, $response->GetStatusCode(), $request, $response);
    }

    return $response;
  }
};

class AcmeException extends RuntimeException
{
};

class AcmeErrorDocException extends AcmeException
{
  public $errorDoc;

  function __construct($errorDoc, $previous=null)
  {
    $this->errorDoc = $errorDoc;
    $message = $errorDoc["type"] . ": " . $errorDoc["detail"];
    if($errorDoc["type"] == "urn:ietf:params:acme:error:userActionRequired" && isset($errorDoc["instance"]))
    {
      $message .= " Visit " . $errorDoc["instance"];
    }
    parent::__construct($message, 0, $previous);
  }
};

abstract Class AcmePostReturn
{
  const JSON = 0;
  const RAW = 1;
};

class AcmeHttpsClient
{
  use Base64Url;

  public $pkey;
  public $kid;
  public $lastResponse;
  public $nonceEndpoint;

  private $httpClient;

  function __construct($verifySSL = true)
  {
    $this->httpClient = new HttpClient([
      "verify" => $verifySSL,
      "redirect" => [
        "retain_post" => true,
      ],
      "http" => false
      ]);
  }
  
  function Get($url)
  {
    try 
    {
      $this->lastResponse = $this->httpClient->Get($url);
    } catch (Exception $ex) 
    {
      $this->HandleException($ex);
    }

    return $this->lastResponse->GetJson();
  }

  function Head($url)
  {
    try
    {
      $this->lastResponse = $this->httpClient->Head($url);
    }
    catch(Exception $ex)
    {
      $this->HandleException($ex);
    }
  }
  
  function Post($url, $payload, $returnFormat = AcmePostReturn::JSON)
  {	
    $protected = [
      "alg" => $this->GetAlgFromKey($this->pkey),
      "url" => $url
    ];
    
    if(isset($this->kid))
    {
      $protected["kid"] = $this->kid;
    }
    else 
    {
      $protected["jwk"] = $this->CreateJwk($this->pkey);
    }    

    $done = false;
    $maxTries = 10;
    while(!$done)
    {
      $protected["nonce"] = $this->GetNonce();
      $data = $this->CreateBody($protected, $payload);
      
      try 
      {
        $this->lastResponse = $this->httpClient->Post($url, [
          "body" => $data,
          "headers" => [
            "Content-Type" => "application/jose+json"
          ]
        ]);
        
        if($returnFormat == AcmePostReturn::JSON)
        {
          return $this->lastResponse->GetJson();
        }
        else 
        {
          return stream_get_contents($this->lastResponse->GetBody(), -1, 0);
        }
      } 
      catch (HttpBadClientResponse $ex) 
      {
        if($maxTries < 1)
        {
          $this->HandleException($ex);
        }

        $this->lastResponse = $ex->GetResponse();
        $errorDoc = $ex->GetResponse()->GetJson();
        
        if($errorDoc === null || $errorDoc["type"] != "urn:ietf:params:acme:error:badNonce")
        {
          $this->HandleException($ex);
        }

        $maxTries -= 1;
      }
      catch(Exception $ex)
      {
        $this->HandleException($ex);
      }
    }
  }

  private function CreateBody($protected, $payload)
  {
    $protected = $this->Base64UrlEncode(json_encode($protected));
    
    if(!is_string($payload))
    {
      $payload = $this->Base64UrlEncode(json_encode($payload));
    }

    $signature = $this->Sign($protected . "." . $payload, $this->pkey);

    return json_encode([
      "protected" => $protected,
      "payload" => $payload,
      "signature" => $signature
    ]);
	}
	
	function ConvertDERtoECDSA($signature, $req_length)
	{
		$separated = [];

		if(ord($signature[0]) != 0x30)
		{
			throw new AcmeException("Invalid DER ECDSA signature compound object");
		}

		$length = ord($signature[1]);

		if(strlen($signature)-2 != $length)
		{
			throw new AcmeException("Invalid DER ECDSA compound length");
		}

		$signature = substr($signature, 2);

		while(strlen($signature) > 0)
		{
			$type = ord($signature[0]);

			if($type != 0x02)
			{
				throw new AcmeException("Invalid DER ECDSA value type");
			}

			$length = ord($signature[1]);

			if($length > $req_length+1)
			{
				throw new AcmeException("Invalid DER ECDSA value length");
			}

			if($length > strlen($signature) -2 )
			{
				throw new AcmeException("DER ECDSA not enough data left in signature");
			}

			$value = substr($signature, 2, $length);
			$signature = substr($signature, 2+$length);

			if($length == $req_length+1)
			{
				$value = substr($value, 1);
			}

			if($length < $req_length)
			{
				$pad = $req_length - $length;
				$value = str_pad($value, $pad, chr(0), STR_PAD_LEFT);
			}

			$separated[] = $value;
		}

		return implode("", $separated);
	}

  function Sign($data, $key)
  {
    $details = PrivateKeyDetails($key);

    switch($details["type"])
    {
      case OPENSSL_KEYTYPE_RSA:
        openssl_sign($data, $signature, $key, "SHA256");
        break;
			case OPENSSL_KEYTYPE_EC:
				switch($details["bits"])
				{
					case 256:
						openssl_sign($data, $signature, $key, "SHA256");
						$signature = $this->ConvertDERtoECDSA($signature, 32);
						break;
					case 384:
						openssl_sign($data, $signature, $key, "SHA384");
						$signature = $this->ConvertDERtoECDSA($signature, 48);
						break;
					default:
						throw new AcmeException("Unsupported EC key curve: " . $details['ec']['curve_name']);
				}
				break;
      default:
        throw new AcmeException("The key type is not supported");
    }

    return $this->Base64UrlEncode($signature);
  }

  function CreateJwk($key)
  {
		$details = PrivateKeyDetails($key);
		
    switch($details["type"])
    {
      case OPENSSL_KEYTYPE_RSA:
        return 
        [
					"e" => $this->Base64UrlEncode($details["rsa"]["e"]),
          "kty" => "RSA",
          "n" => $this->Base64UrlEncode($details["rsa"]["n"])
        ];
			case OPENSSL_KEYTYPE_EC:
				return
				[
					"crv" => "P-" . $details['bits'],
					"kty" => "EC",
					"x" => $this->Base64UrlEncode($details["ec"]["x"]),
					"y" => $this->Base64UrlEncode($details["ec"]["y"]),
				];
      default:
        throw new AcmeException("The key type is not supported");
    }
  }

  private function GetNonce()
  {
    if(!isset($this->nonceEndpoint))
    {
      throw new AcmeException("The ACME nonce endpoint has not been set");
    }

    if(!isset($this->lastResponse) || $this->lastResponse->GetHeader("replay-nonce") === null)
    {
      $this->Head($this->nonceEndpoint);
    }

    $nonce = $this->lastResponse->GetHeader('replay-nonce')[0];
    
    if($nonce === null)
    {
      throw new AcmeException("The server did not supply a replay nonce");
    }
    
    return $nonce;
  }


  private function HandleException($ex)
  {
    if($ex instanceof HttpBadClientResponse)
    {
      $this->lastResponse = $ex->GetResponse();
      $json = $this->lastResponse->GetJson();

      if($json)
      {
        throw new AcmeErrorDocException($json, $ex);
      }
    }
    else 
    {
      throw $ex;	
    }
  }

  private function GetAlgFromKey($key)
  {
    $details = PrivateKeyDetails($key);

    switch($details["type"])
    {
      case OPENSSL_KEYTYPE_RSA:
        return "RS256";
			case OPENSSL_KEYTYPE_EC:
				switch($details['bits'])
				{
					case 256:
					case 384:
						return "ES" . $details['bits'];
					case 521:
						return "ES512";
				}
      default:
        throw new AcmeException("The private key type is not supported");
    }
  }
}

class AcmeClientUnprotected
{
  use Base64Url;

  public $baseUrl;
  public $accountUrl;

  private $httpClient;
  private $nonce;
  private $config;

  function __construct($baseUrl, $verifySSL = true)
  {
    $this->baseUrl = $baseUrl;
    $this->httpClient = new AcmeHttpsClient($verifySSL);
    $this->Configure();
  }

  private function Configure()
  {
    $this->config = $this->httpClient->Get($this->baseUrl);

    if($this->config == null)
    {
      throw new AcmeException("Could not get any configuration from ACME server");
    }

    if(!$this->HasAllKeys($this->config, [
      "newNonce",
      "newAccount",
      "newOrder",
      "revokeCert",
      "keyChange"
    ]))
    {
      throw new AcmeException("The server did not supply all required configuration fields");
    }
    
    $this->httpClient->nonceEndpoint = $this->config["newNonce"];
  }
  
  private function CreateCsr($domains, $keyConfig)
  {
		//var_dump($domains);
		//var_dump($keyConfig);
		//var_dump($_POST);
		//exit;
	
    $san = '';

    foreach($domains as $domain)
    {
      if(strlen($san) > 0) $san .= ',';
      $san .= strtoupper($domain['type']) . ':' . $domain['value'];
    }

    $commonName = $domains[0]["value"];

    $csr = CreateCsr(
      [
        "key" => $keyConfig,
        "extensions" => ["subjectAltName" => $san],
        "distinguishedName" => ["commonName" => $commonName]
      ]);

    $pem = ExportCsr($csr["csr"]);
    preg_match('~REQUEST-----(.*)-----END~s', $pem, $matches);
    
    return [
      "key" => $csr["key"],
      "csr" => $this->Base64UrlEncode(base64_decode($matches[1]))
    ];
  }

  private function HasAllKeys(array $in, $required_keys)
  {
    return !array_diff_key(array_flip($required_keys), $in);
  }

  private function PerformChallenge($challenge)
  {
    $dir = "/var/apresa/acme-challenge/";
    $filename = $dir . $challenge["token"];

    $details = PrivateKeyDetails($this->httpClient->pkey);
    $data = $this->httpClient->CreateJwk($this->httpClient->pkey);

    $thumbprint = $this->Base64UrlEncode(hash("sha256", json_encode($data), true));
    $key = $challenge["token"] . "." . $thumbprint;

    if(file_put_contents($filename, $key) === false)
    {
      throw new AcmeException("Could not create file for http challenge");
    }

    $this->httpClient->Post($challenge["url"], (object) []);

    return $filename;
  }

  private function ThrowOnInvalidAuth($authorization)
  {
    foreach($authorization["challenges"] as $challenge)
    {
      if($challenge["status"] == "invalid" && isset($challenge["error"]))
      {
        throw new AcmeErrorDocException($challenge["error"]);
      }
    }

    throw new AcmeException("The authorization is invalid");
  }

  private function PollAuth($authurl)
  {
    $polInterval = 2;
    $pending = true;
    $maxTries = 30;

    $authorization = [];

    while($pending)
    {
      $authorization = $this->httpClient->Post($authurl, "");

      if($authorization["status"] != "pending")
      {
        $pending = false;
      }
      else 
      {
        $maxTries--;

        if($maxTries < 1)
        {
          throw new AcmeException("Reached polling limit on authorization object");
        }

        sleep($polInterval);
      }
    }
    
    return $authorization;
  }

  private function PerformAuth($order)
  {
    try
    {
		  // Permission issue, so ask the handler to create the folder structure
		  $msg = [];
		  if(!msg_signal_service($msg, "acmechallengecreate"))
      {
        throw new AcmeException("Could not create directory structure required: " . end($msg)->text);
      }

      foreach($order["authorizations"] as $authurl)
      {
        $authorization = $this->httpClient->Post($authurl, "");
        
        if($authorization["status"] == "valid") continue;

        switch($authorization["status"])
        {
          case "revoked":
            throw new AcmeException("Authorization has been revoked by the server");
          case "deactivated":
            throw new AcmeException("Authorization has been deactivated by the client");
          case "invalid":
            $this->ThrowOnInvalidAuth($authorization);
          case "expired":
            throw new AcmeException("Authorization has expired");
        }

        if($authorization["status"] == "pending")
        {
          foreach($authorization["challenges"] as $challenge)
          {
            if($challenge["type"] == "http-01")
            {
              $theChallenge = $challenge;
            }
          }

          if(!isset($theChallenge))
          {
            throw new AcmeException("Server did not return an http-01 challenge. Did you request a wildcard certificate?");
          }

          try
          {
            $filename = $this->PerformChallenge($theChallenge);
            $authorization = $this->PollAuth($authurl);
          }
          finally
          {
            unlink($filename);
          }
          
          $theChallenge = null;
        }
        
        switch($authorization["status"])
        {
          case "revoked":
            throw new AcmeException("Authorization has been revoked by the server");
          case "deactivated":
            throw new AcmeException("Authorization has been deactivated by the client");
          case "invalid":
            $this->ThrowOnInvalidAuth($authorization);
          case "expired":
            throw new AcmeException("Authorization has expired");
        }
      }
    }
    finally
    {
      $msg = [];
      msg_signal_service($msg, "acmechallengedelete");
    }
  }

  private function PollOrder($orderUrl, $state)
  {
    $processing = true;
    $pollInterval = 2;
    $maxTries = 30;

    $order = [];

    while($processing) 
    {	
      $order = $this->httpClient->Post($orderUrl, "");

      if($order["status"] == $state)
      {
        $processing = false;
      }
      else if($order["status"]  == "invalid")
      {
        $this->ThrowOnInvalidOrder($order);
      }
      else 
      {
        $maxTries--;

        if($maxTries < 1)
        {
          throw new AcmeException("Reached polling limit on authorization object when waiting for state $state");
        }

        sleep($pollInterval);
      }
    }

    return $order;
  }

  private function SplitCerts($certs)
  {
    $out = [];

    $nrMatches = preg_match_all('/(-{5}BEGIN CERTIFICATE-{5}(?:\n|\r\n)(?:[a-zA-Z0-9\+\/]+(?:\n|\r\n){0,1})+={0,2}(?:\n|\r\n)-{5}END CERTIFICATE-{5})/', $certs, $matches, PREG_SET_ORDER);

    if(!$nrMatches)
    {
      throw new Exception("Could not extract certificates returned by server");
    }

    foreach($matches as $cert)
    {
      $out[] = ImportCertificate($cert[0]);
    }

    return $out;
  }

  private function DownloadCert($order)
  {
    $certs = $this->httpClient->Post($order["certificate"], "", AcmePostReturn::RAW);
    return $this->SplitCerts($certs);
  }

  private function ThrowOnInvalidOrder($order)
  {
    if(isset($order["error"]))
    {
      throw new AcmeErrorDocException($order["error"]);
    }
    
    throw new AcmeException("The order is invalid");
  }
  
  function RegisterAccount($email)
  {
    $pkey = CreatePrivateKey();

    $this->httpClient->pkey = $pkey;
    $this->httpClient->kid = null;

    $payload = [
     "termsOfServiceAgreed" => true,
     "contact" => [
        "mailto:" . $email
       ]
    ];
    
    $account = $this->httpClient->Post($this->config['newAccount'], $payload);
    $this->accountUrl = $this->httpClient->lastResponse->GetHeader("location")[0];

    if($this->accountUrl === null)
    {
      throw new AcmeException("The server did not return an account url");
    }

    $this->httpClient->kid = $this->accountUrl;

    return [
      "key" => $pkey,
      "url" => $this->accountUrl
    ];
  }
  
  function UpdateAccount($email)
  {
    $payload = [
      "contact" => [
        "mailto:" . $email
      ]
    ];
    
    $this->httpClient->Post($this->accountUrl, $payload);
  }

  function DeactivateAccount()
  {
    $payload = [
      "status" => "deactivated"
    ];
    
    $this->httpClient->Post($this->accountUrl, $payload);
  }
  
  function RequestCert($domains, $keyConfig)
  {
    $csr = $this->CreateCsr($domains, $keyConfig);

    $payload = [ "identifiers" => $domains ];

    $order = $this->httpClient->Post($this->config['newOrder'], $payload);
    $orderUrl = $this->httpClient->lastResponse->GetHeader("location")[0]; 
    
    if($order["status"] == "pending")
    {
      $this->PerformAuth($order);
    }

    $order = $this->PollOrder($orderUrl, "ready");
    $this->httpClient->Post($order['finalize'], ["csr" => $csr["csr"]]);
    $order = $this->PollOrder($orderUrl, "valid");

    $certificates = $this->DownloadCert($order);
    return [
      "certs" => $certificates,
      "key" => $csr["key"]
    ];
  }
  
  function RevokeCert($cert, $reason=0)
  {
    $pem = ExportCert($cert);
    $res = preg_match('/-{5}BEGIN CERTIFICATE-{5}\n((?:[a-zA-Z0-9\+\/]+\n{0,1})+={0,2})\n-{5}END CERTIFICATE-{5}/', $pem, $matches);

    if(!$res)
    {
      throw new AcmeException("Failed to process certificates for revocation request");
    }

    $base64urlCert = $this->Base64UrlEncode(base64_decode($matches[1]));
		
		$payload = ["certificate" => $base64urlCert];
		if($reason != 0)
		{
			$payload["reason"] = $reason;
		}

    $this->httpClient->Post($this->config["revokeCert"], $payload);
  }

  function SetAccount(array $account)
  {
    $this->accountUrl = $account["url"];
    $this->httpClient->kid = $account["url"];
    $this->httpClient->pkey = $account["key"];
  }

	function GetSubscriberAgreement()
	{
		if(isset($this->config['meta']['termsOfService']))
		{
			return $this->config['meta']['termsOfService'];
		}
		else
		{
			return 'data:text/plain,Unavailable';
		}
	}
};

class AcmeClient
{
  private $client;
  private $fp;
  private $locked;

  function __construct($baseUrl, $verifySSL = true)
  {
    $this->client = new AcmeClientUnprotected($baseUrl, $verifySSL);
  }

	function GetSubscriberAgreement()
	{
		return $this->client->GetSubscriberAgreement();
	}

  function Lock()
  {
		@mkdir("/tmp/acme");
		@chmod("/tmp/acme", 0777);
    
		$this->fp = fopen("/tmp/acme/acmelock", "a+");
		
		if(!flock($this->fp, LOCK_EX))
    {
      throw new AcmeException("Could not get exclusive access");
		}
		
		chmod("/tmp/acme/acmelock", 0666);
    $this->locked = true;
  }
  
  function Unlock()
  {
    if($this->locked)
    {
      flock($this->fp, LOCK_UN);
    }

    if($this->fp)
    {
      fclose($this->fp);
    }
  }

  function SetAccount($account)
  {
    return $this->client->SetAccount($account);
  }

  function RegisterAccount($email)
  {
    try
    {
      $this->Lock();
      return $this->client->RegisterAccount($email);
    }
    finally
    {
      $this->Unlock();
    }
  }

  function UpdateAccount($email)
  {
    try
    {
      $this->Lock();
      return $this->client->UpdateAccount($email);
    }
    finally
    {
      $this->Unlock();
    }
  }

  function DeactivateAccount()
  {
    try
    {
      $this->Lock();
      $this->client->DeactivateAccount();
    }
    finally
    {
      $this->Unlock();
    }
  }

  function RequestCert($domains, $keyConfig)
  {
    try
    {
      $this->Lock();
      return $this->client->RequestCert($domains, $keyConfig);
    }
    finally
    {
      $this->Unlock();
    }
  }

  function RevokeCert($cert, $reason=0)
  {
    try
    {
      $this->Lock();
      $this->client->RevokeCert($cert, $reason);
    }
    finally
    {
      $this->Unlock();
    }
  }
};
?>