<?php
// Utility file containg features related to Azure

use Microsoft\Graph\Graph;
use Microsoft\Graph\Model;

require_once dirname(__FILE__) . '/../apresa/vendor/autoload.php';
require_once dirname(__FILE__) . '/actions.php';

function GetAzureAccessToken($azid, & $errmsg)
{
  global $dbintf;
  try
  {
    // get the application details
    $app = $dbintf->GetAzureApp($azid);
    if ($app===false)
    {
      $errmsg = 'Azure app not found in database';
      return false;
    }
    
    $guzzle = new \GuzzleHttp\Client();
    
    $clientId = $app['appid'];
    $clientSecret = decrypt( $app['appsecret'] );
    $tenantId = $app['mstenant'];
    
    $url = 'https://login.microsoftonline.com/' . $tenantId . '/oauth2/token?api-version=1.0';
    
    $response = $guzzle->post($url, [
      'form_params' => [
          'client_id' => $clientId,
          'client_secret' => $clientSecret,
          'resource' => 'https://graph.microsoft.com/',
          'grant_type' => 'client_credentials',
      ]
    ]);
    $token = json_decode( $response->getBody()->getContents() );
    $accessToken = $token->access_token;
  }
  catch (Exception $e)
  {
    $errmsg = $e->getMessage();
    return false;
  }
  return $accessToken;
}

function GetAzureGroupList($azid, & $errmsg)
{
  $accessToken = GetAzureAccessToken($azid, $errmsg);
  if ($accessToken===false)
  {
    return null;
  }
  
  $graph = new Microsoft\Graph\Graph();
  $graph->setAccessToken($accessToken);
  
  $groupWalk = $graph->createCollectionRequest("GET", "/groups")
                ->setReturnType(Model\Group::class)
                ->setPageSize(5);

  $list = array();
  while ($groupWalk->isEnd()==false)
  {
    $groups = $groupWalk->getPage();
    foreach ($groups as $group)
    {
      $list[] = (object) [ 'id' => $group->getId(), 'name' => $group->getDisplayName() ];
    }
  }
  return $list;
}

function AzureSyncGroup($group)
{
	try
	{
		return AzureSyncGroupE($group);
	} catch (Exception $e)
	{
		lg('Azure group sync error: ' . $e->getMessage() );
		return false;
	}
}

function azThrow($msg)
{
	throw new Exception($msg);
}

function AzureSyncGroupE($group)
{
  global $dbintf;
  
  if ($group['azuregroupid']=='') azThrow('Unconfigured');
  if ($group['azureapp']==null) azThrow('Unconfigured');

  // get the application details
  $app = $dbintf->GetAzureApp($group['azureapp']);
  if ($app===false) azThrow('Azure app configuration not in database');
  
  $errmsg = '';
  $accessToken = GetAzureAccessToken( $group['azureapp'], $errmsg );
  if ($accessToken===false)
  {
		azThrow('Could not get Azure access token: ' . $errmsg);
  }
  $azureapp = (int) $group['azureapp'];
  
  lg('Synchronizing group ' . $group['name'] . ' with Azure AD group ' 
    . $group['azuregroupname'] . ' with id ' . $group['azuregroupid'] );
    
  $graph = new Microsoft\Graph\Graph();
  $graph->setAccessToken($accessToken);
  
  $userWalk = $graph->createCollectionRequest("GET", "/groups/" . $group['azuregroupid'] . '/members')
              ->setReturnType(Model\User::class)
              ->setPageSize(5);
  $adUsers = [];
  while ($userWalk->isEnd()==false)
  {
    $users = $userWalk->getPage();
    foreach ($users as $user)
    {
      //lg($user->getUserPrincipalName());
      $adUser = [];
      $adUser['upn'] = $user->getUserPrincipalName();
      $adUser['name'] = $user->getDisplayName();
      $adUser['businessPhones'] = $user->getBusinessPhones();
      $adUser['mobilePhone'] = $user->getMobilePhone();
      $adUser['email'] = $user->getMail();
      $adUsers[] = $adUser;
    }
  }
  
  $members = [];
  
  // import each user we have found
  foreach ($adUsers as $adUser)
  {
    // use UPN as username
    $usrdata = array();
    $usrdata['usr'] = $adUser['upn'];
    $usrdata['name'] = $adUser['name'];
    $usrdata['email'] = $adUser['email'];
    if ($usrdata['email']===null) $usrdata['email'] = '';
    
    if ($dbintf->UsrExists($usrdata['usr'])==false)
    {
      // this is a new user
      // set defaults
      $usrdata['mphones'] = '';
      $usrdata['enabled'] = 1;
      $usrdata['rod'] = 0;
      $usrdata['logonmethod'] = 2; // using external identity provider
      $usrdata['playtimelim'] = 0;
      $usrdata['ldapuser'] = 0;
      $usrdata['ldaphost'] = '';
      $usrdata['samwinuser'] = 0;
      $usrdata['azureapp'] = $azureapp;
      lg('Adding user: '.$usrdata['usr'].' ('.$usrdata['name'].')');
      $isok = $dbintf->AddUserAccount($userid,$usrdata,0);
      $accid = $dbintf->GetUserInfo($userid)['accid'];
    } else
    {
      // existing user
      // verbose: lg($usrdata['usr'] . ' is an existing user');
      // the user already exists
      // get existing info
      $existing = $dbintf->GetUserInfoOnUsername($usrdata['usr']);
      // update user?
      if ($existing['email'] != $usrdata['email']
      || $existing['name'] != $usrdata['name'])
      {
        lg('Editing user: '.$usrdata['usr'].' ('.$usrdata['name'].')');
        $dbintf->EditUserAccountLC($existing['id'],$usrdata);
      }
      $accid = $existing['accid'];
    }
		$members[] = $accid;
    
    // the phones numbers or the equiv
    $phone_nums = array();
    
    // UPN as phone
    if ($app['importphones_upn']>0)
    {
      $phone_nums[] = $adUser['upn'];
    }
    if ($app['importphones_bphones']>0)
    {
      $phone_nums = array_merge($phone_nums , $adUser['businessPhones'] );
    }
    if ($app['importphones_name']>0)
    {
      $phone_nums[] = $adUser['name'];
    }
    if ($app['importphones_mobile']>0)
    {
      if ($adUser['mobilePhone']!=null)
      {
        $phone_nums[] = $adUser['mobilePhone'];
      }
    }
    
    // convert the phone numbers to contacts
    $phones = array();
    $newphones = array();
    foreach ($phone_nums as $num)
    {
      // fill in name
      $res = $dbintf->GetOrCreateCallerID($num,'',1,$usrdata['name']);
      $phones[] = array('contactid'=>$res['contactid']);
      $newphones[] = $res['contactid'];
    }		  

    $oldphones = $dbintf->GetUserPhones($accid);
    if (count(array_diff($newphones,$oldphones))!=0
    || count(array_diff($oldphones,$newphones))!=0)
    {
      lg('Phones list changed for user: '.$usrdata['usr']);
      // the list of phones: implode(',',$phone_nums));	
      $dbintf->SetUserPhones($accid,$phones);
    }    
    
  } // looping through AD users
  
  $oldmembers = $dbintf->GetGroupMembers($group['accid']);
  // Are there any useraccounts made by samwin sync and not part of any group anymore?
  $delusers = array_diff($oldmembers, $members);
  // SetGroupMembers is quite slow!!!
  if (count($delusers) != 0 || (count($oldmembers) != count($members)))
  {
    lg('The group now has '.count($members).' members');
    $dbintf->SetGroupMembers($group['accid'],$members);
  } else
  {
    lg('Group membership unchanged');
  }
  
  foreach ($delusers as $deluser)
  {
    $user = $dbintf->GetUserInfoOnAccId($deluser);
    lg('Removing '.$user['usr'].' from group');
    // check if user was created by this same azure app
    if ($user['azureapp']==$azureapp)
    {
      $dbgroups = $dbintf->GetGroupsOfWhichMember($deluser);
      lg('It is still member of '.count($dbgroups).' groups');
      if (count($dbgroups)==0)
      {
        lg("Removing user: ".$user['usr']);
        //The user account was created by Samwin sync,
        // and is no longer member of any group => delete
        $dbintf->DeleteUserAccount($user['id']);
      }
    }
  }

  
  return true;
}



