<?php
	require_once("utility.php");
	require_once("acmeclient.php");

  define("CERTDIR", "/var/apresa/certificates/");
  define("ACME_ENDPOINT", "https://acme-v02.api.letsencrypt.org/directory");
  define("ACME_VERIFY_SSL", true);

  abstract class OPENSSL_CONFIG_FILE
  {
    const certificate = 0;
    CONST csr = 1;
	}

  function PrivateKeyDetails($key)
  {
    $details = openssl_pkey_get_details($key);

    if($details === null)
    {
      throw new Exception("Could not get private key details: " . openssl_error_string());
    }

    return $details;
  }
  
  function CreatePrivateKey($config = [])
  {
    $pkey = openssl_pkey_new($config);

    if($pkey === false)
    {
      throw new Exception("Could not create private key: " . openssl_error_string());
    }

    return $pkey;
  }

  function ExportKey($key)
  {
    if(!openssl_pkey_export($key, $pem))
    {
      throw new Exception("Could not export private key: " . openssl_error_string());
    }

    return $pem;
  }

  function ExportKeyToFile($key, $filename)
  {
    if(!openssl_pkey_export_to_file($key, $filename))
    {
      throw new Exception("Could not export private key to disk: " . openssl_error_string());
    }
  }

  function ImportKey($PEM)
  {
    $key = openssl_pkey_get_private($PEM);
    if($key === false)
    {
      throw new Exception("Could not import private key: " . openssl_error_string());
    }

    return $key;
  }

  function ExportCsr($csr)
  {
    if(!openssl_csr_export($csr, $pem))
    {
      throw new Exception("Could not export the signing request: " . openssl_error_string());
    }

    return $pem;
  }

  function ExportCsrToFile($csr, $file)
  {
    if(!openssl_csr_export_to_file($csr, $file))
    {
      throw new Exception("Could not export the signing request to disk: " . openssl_error_string());
    }
  }

  function ExportCertToFile($cert, $filename)
  {
    if(!openssl_x509_export_to_file($cert, $filename))
    {
      throw new Exception("Could not export certificate to disk: " . openssl_error_string());
    }
  }

  function ExportCertBundleToFile($certs, $filename)
  {
    $pemCerts = [];

    foreach($certs as $cert)
    {
      $pemCerts[] = ExportCert($cert);
    }

    $out = implode("\n", $pemCerts) . "\n";

    if(file_put_contents($filename, $out) === false)
    {
      throw new Exception("Could not write certificate bundle to disk");
    }
  }

  function ImportCertificate($cert)
  {
    $cert = openssl_x509_read($cert);

    if($cert === false)
    {
      throw new Exception("Could not import the certificate: " . openssl_error_string());
    }

    return $cert;
  }

  function ExportCert($cert)
  {
    if(!openssl_x509_export($cert, $pem))
    {
      throw new Exception("Could not export the certificate: " . openssl_error_string());
    }

    return $pem;
  }

  function CertificateDetails($cert)
  {
    $details = openssl_x509_parse($cert);

    if($details === false)
    {
      throw new Exception("Could not get certificate details: " . openssl_error_string());
    }

    return $details;
  }

  function CreateConfigFile($extensions, $configType)
  {
    $filename = "/tmp/" . uniqid() . ".cnf";
    $config_file = fopen($filename, 'w');

    if($config_file === false)
    {
      throw new Exception("Could not write certificate configuration to disk");
    }
        
    fwrite($config_file, "[ req ]\n");
    fwrite($config_file, "distinguished_name = req_distinguished_name\n");
    if($configType == OPENSSL_CONFIG_FILE::csr) fwrite($config_file, "req_extensions = v3_ca\n");
    fwrite($config_file, "prompt = no\n");
    fwrite($config_file, "[ req_distinguished_name ]\n");
    fwrite($config_file, "[ v3_ca ]\n");
    fwrite($config_file, "subjectKeyIdentifier = hash\n");
    if($configType == OPENSSL_CONFIG_FILE::certificate) fwrite($config_file, "authorityKeyIdentifier=keyid:always,issuer\n");
    
    foreach($extensions as $extension=>$value)
    {
      $line = "$extension = $value\n";
      fwrite($config_file, $line);
    }

    fclose($config_file);

    return $filename;
  }
  
  function CreateCsr($configs, $privKeyConfigs=null)
  {
		$key = CreatePrivateKey($configs["key"]);

    $configFile = CreateConfigFile($configs["extensions"], OPENSSL_CONFIG_FILE::csr);

    $sslConfig = array(
      "digest_alg" => "sha256",
      "config" => $configFile
    );

    $csr = openssl_csr_new($configs["distinguishedName"], $key, $sslConfig);

    unlink($configFile);

    if($csr === false)
    {
      throw new Exception("Could not create certificate signing request");
    }

    return [
      "key" => $key,
      "csr" => $csr
    ];
  }
  
  function GetCertificateInfoString($filename)
  {
    $cert_file = CERTDIR . $filename;
    $command = "openssl crl2pkcs7 -nocrl -certfile " . escapeshellarg($cert_file) . " | openssl pkcs7 -print_certs -text -noout";
    $info = "";
    exec($command, $info);
    return implode("\n", $info);
  }
  
  function GetCertificateRequestInfoString($filename)
  {
    $cert_file = CERTDIR . $filename;
    $command = "openssl req -noout -text -in " . escapeshellarg($cert_file);
    $info = "";
    exec($command, $info);
    return implode("\n", $info);
  }
  
  function DownloadCertificate($localname, $downloadname)
  {
    $cert_file = CERTDIR . $localname;
    header('Content-Description: File Transfer');
    header('Content-Type: text/plain');
    header('Content-Disposition: attachment; filename=' . $downloadname); 
    header('Pragma: '); // needed for IE (HTTPS)
    ReadFile($cert_file);
    exit;
  }

  function SplitCerts($certs)
  {
    $out = [];

    $nrMatches = preg_match_all('/(-{5}BEGIN CERTIFICATE-{5}(?:\n|\r\n)(?:[a-zA-Z0-9\+\/]+(?:\n|\r\n){0,1})+={0,2}(?:\n|\r\n)-{5}END CERTIFICATE-{5})/', $certs, $matches, PREG_SET_ORDER);

    if(!$nrMatches)
    {
      throw new Exception("Could not extract any certificates from bundle");
    }

    foreach($matches as $cert)
    {
      $out[] = ImportCertificate($cert[0]);
    }

    return $out;
  }

  function UploadCertificate($name, $certificates)
  {
    global $msg, $dbintf;

    $baseFileName = $baseFileName = uniqid($dbintf->actuinfo['id'], true);
    $toDB = [
      "name" => $name,
      "trusted" => 0
    ];

    $cert = ImportCertificate($certificates["cert"]);
    $details = CertificateDetails($cert);

    if(isset($certificates["key"]))
    {
      $pkey = ImportKey($certificates["key"]);
      if(openssl_x509_check_private_key($cert, $pkey) === false)
      {
        throw new Exception("Private key does not match with certificate: " . openssl_error_string());
      }
    }

    if(isset($certificates["intermediate"]))
    {
      $intermediate = SplitCerts($certificates["intermediate"]);
    }

    $toDB["cert_location"] = $baseFileName . ".crt";
    ExportCertToFile($cert, CERTDIR . $toDB["cert_location"]);

    if(isset($pkey))
    {
      $toDB["key_location"] = $baseFileName . ".key";
      $toDB["privatekey"] = 1;
      ExportKeyToFile($pkey, CERTDIR . $toDB["key_location"]);
      msg_signal_service($msg, "pkeypermissions root " . $toDB['key_location']);
    }

    if(isset($intermediate))
    {
      $toDB["intermediate_location"] = $baseFileName . "-CA.crt";
      ExportCertBundleToFile($intermediate, CERTDIR . $toDB["intermediate_location"]);
    }

    $toDB['subject'] = str_ensure_defined($details['subject']['CN']);
    $toDB['valid_to'] = int_ensure_defined($details['validTo_time_t']);

    $id = $dbintf->InsertCertificate($toDB);

    if($id === false)
    {
      throw new Exception("Could not store certificate information in database");
    }

    return $id;
  }
  
  function DeleteCertificate($id)
  {
    global $dbintf, $msg, $lng;
    
    $cert = $dbintf->GetCertificate($id);
        
    if($cert === false)
    {
      $msg[] = new Msg(MSG_ERR, 'Could not retrieve certificate information from database');
      return;
    }
        
    $inuse = $dbintf->CertificateInUse($id);
        
    if($inuse)
    {
      $msg[] = new Msg(MSG_ERR, $cert['name'] . ': ' . $lng->Get('Crypt.DelCertInUse'));
      return;
    }
        
    $deleted = $dbintf->DeleteCertificate($id);
        
    if($deleted === false)
    {
      $msg[] = new Msg(MSG_ERR, $cert['name'] . ': Could not delete certificate information from database');
      return;
    }
        
    if($cert['trusted'])
    {
      msg_signal_service($msg, 'xcertnottrust ' . $cert['cert_location']);
    }
        
    LogAction("Delete certificate",array('id'=>$id,'name'=>$cert['name']));
    
    $certfile = CERTDIR . $cert['cert_location'];
    @unlink($certfile);
        
    if($cert['privatekey'])
    {
      $certkey = CERTDIR . $cert['key_location'];
      @unlink($certkey);
    }
        
    if($cert['intermediate_location'] != '')
    {
      $interfile = CERTDIR . $cert['intermediate_location'];
      @unlink($interfile);
    }
  }
  
  function DeleteCertificateRequest($id)
  {
    global $dbintf, $msg;
    
    $req = $dbintf->GetCertificateRequest($id);
        
    if($req === false)
    {
      $msg[] = new Msg(MSG_ERR, 'Could not retrieve certificate signing request information from database');
      return false;
    }
        
    $deleted = $dbintf->DeleteCertificateRequest($id);
        
    if($deleted === false)
    {
      $msg[] = new Msg(MSG_ERR, $cert['name'] . ': Could not delete certificate signing request information from database');
      return false;
    }

    LogAction("Delete certificate signing request",array('id'=>$id,'name'=>$req['name']));
    
    @unlink(CERTDIR . $req['request_location']);
    @unlink(CERTDIR . $req['key_location']);
    
    return true;
  }
  
  function SelfSignCertificate($configsIn, $daysValid)
  {
    $csr = CreateCsr($configsIn);

    $configFile = CreateConfigFile($configsIn["extensions"], OPENSSL_CONFIG_FILE::certificate);
    $sslConfig = array(
      "digest_alg" => "sha256",
      "config" => $configFile
    );

    $sslConfig["x509_extensions"] = "v3_ca";
    $crt = openssl_csr_sign($csr["csr"], NULL, $csr["key"], $daysValid, $sslConfig, rand());
    unlink($configFile);

    if($crt === false)
    {
      throw new Exception('Could not sign certificate: ' . openssl_error_string());
    }

    return ["cert" => $crt, "key" => $csr["key"]];
	}

	function GetLetsEncryptSubscriberAgreement()
	{
		$client = new AcmeClient(ACME_ENDPOINT, ACME_VERIFY_SSL);
		return $client->GetSubscriberAgreement();
	}
	
	function RenewLetsEncrypt($id)
	{
    global $dbintf;
    $msg = [];

    $certDB = $dbintf->GetCertificate($id);
    
		$certFile = "file://" . CERTDIR . $certDB['cert_location'];
		$cert = ImportCertificate($certFile);
    $certDetails = CertificateDetails($cert);
    $keyFile = "file://" . CERTDIR . $certDB['key_location'];
    $key = ImportKey($keyFile);

		$certDomains = explode(",", $certDetails["extensions"]["subjectAltName"]);

		$domains = [];

		foreach($certDomains as $domain)
		{
			$domain = trim($domain);
			$split = explode(":", $domain);
			if(sizeof($split) != 2) continue;
			if(strtolower($split[0]) != "dns") continue;

			$domains[] = [
				"type" => "dns",
				"value" => $split[1]
			];
    }
    
    $keyDetails = PrivateKeyDetails($key);

    switch($keyDetails['type'])
    {
      case OPENSSL_KEYTYPE_RSA:
        $keyConfigs = [
          "private_key_type" => OPENSSL_KEYTYPE_RSA,
          "private_key_bits" => $keyDetails['bits']
        ];
      break;
      case OPENSSL_KEYTYPE_EC:
        $keyConfigs = [
          "private_key_type" => OPENSSL_KEYTYPE_EC,
          "curve_name" => $keyDetails["ec"]["curve_name"]
        ];
      break;
      default:
        throw new Exception("Unsupported private key type");
    }
		
		$client = new AcmeClient(ACME_ENDPOINT, ACME_VERIFY_SSL);
		$client->SetAccount(GetLetsEncryptAccount());

		$certs = $client->RequestCert($domains, $keyConfigs);

		$baseName = uniqid(ensure_defined($dbintf->actuinfo['id'], 'le'), true);
		$certFile = $baseName . '.crt';
    $certKey = $baseName . '.key';
    
    $newCert = array_shift($certs["certs"]);
		
		ExportKeyToFile($certs["key"], CERTDIR . $certKey);
		ExportCertToFile($newCert, CERTDIR . $certFile);

		if(sizeof($certs["certs"]) > 0)
		{
			$intermediateFile = $baseName . '-CA.crt';
			ExportCertBundleToFile($certs["certs"], CERTDIR . $intermediateFile);
		}
		else 
		{
			$intermediateFile = null;
		}

    $certDetails = openssl_x509_parse($newCert);
    
		$toDB = [
			"subject" => str_ensure_defined($certDetails['subject']['CN']),
			"valid_to" => int_ensure_defined($certDetails['validTo_time_t']),
			"cert_location" => $certFile,
			"key_location" => $certKey,
			"intermediate_location" => $intermediateFile,
			"lets_encrypt" => 1
		];

		$dbintf->UpdateCertificate($id, $toDB);
		
		if(!rename(CERTDIR . $certDB['cert_location'], CERTDIR . "/lets_encrypt_old/" . $certDB['cert_location']))
		{
			unlink(CERTDIR . $certDB['cert_location']);
		}

		if(!rename(CERTDIR . $certDB['key_location'], CERTDIR . "/lets_encrypt_old/" . $certDB['key_location']))
		{
			unlink(CERTDIR . $certDB['key_location']);
		}

		if(isset($certDB['intermediate_location'])) 
		{
			if(!rename(CERTDIR . $certDB['intermediate_location'], CERTDIR . "/lets_encrypt_old/" . $certDB['intermediate_location']))
			{
				unlink(CERTDIR . $certDB['intermediate_location']);
			}
		}
	}

	function RevokeLetsEncrypt($id, $reason=0)
	{
		global $dbintf;
		$cert = $dbintf->GetCertificate($id);
        
		$filestr = "file://" . CERTDIR . $cert['key_location'];
		$key = ImportKey($filestr);
		$filestr = "file://" . CERTDIR . $cert['cert_location'];
		$cert = ImportCertificate($filestr);

		// Note, two possibilities to revoke a certificate:
		// Sign messages to acme server with account key for which cert was issued, or with the private key of the cert itself
		// Here we use the key of the cert itself, as a new account may have been created in the meantime.
		
		$account = ["key" => $key, "url" => null];
		$client = new AcmeClient(ACME_ENDPOINT, ACME_VERIFY_SSL);
		$client->SetAccount($account);

		if($reason != 1 && $reason != 4 && $reason != 5)
		{
			$reason = 0;
		}

		$client->RevokeCert($cert, $reason);

		$todb = array(
			'lets_encrypt' => 2
		);

		$dbintf->UpdateCertificate($id, $todb);
	}

	function TransfromWebPrivKey($configs)
	{
		if(PHP_VERSION_ID < 70100)
		{
			// No support for creation of elliptic curve keys, so always force RSA
			$configs["KeyType"] = "RSA";
		}

		$out = [];
		switch(str_ensure_defined($configs["KeyType"]))
		{
			case "EC":
				$out["private_key_type"] = OPENSSL_KEYTYPE_EC;
				$out["curve_name"] = isset($configs["ECCurve"]) ? $configs["ECCurve"] : "prime256v1";
				break;
			case "RSA":
			default:
				$out["private_key_type"] = OPENSSL_KEYTYPE_RSA;
				$out["private_key_bits"] = isset($configs["RSABits"]) ? intval($configs["RSABits"]) : 2048;
				break;
		}

		return $out;
  }

  function TransfromWebDistinguishedName($fromWeb)
  {	
    if(!isset($fromWeb['CertName']))
    {
      throw new Exception("Common name is required");
    }
    
    $webToCert = array(
      "CertName" => "commonName",
      "CertCountry" => "countryName",
      "CertState" => "stateOrProvinceName",
      "CertCity" => "localityName",
      "CertOrg" => "organizationName",
      "CertOrgUnit" => "organizationalUnitName",
      "CertEmail" => "emailAddress"
    );
    
    $distinguishedName = array();
    
    foreach($webToCert as $webName => $certName)
    {
      if(str_ensure_defined($fromWeb[$webName]) != '')
      {
        $distinguishedName[$certName] = $fromWeb[$webName];
      }
    }
    
    return $distinguishedName;
  }
  
  function ToConfString($fromWeb, $keyMapping)
  {
    $conf = [];
    foreach($keyMapping as $key=>$mapped)
    {
      if(str_ensure_defined($fromWeb[$key]) === "1")
      {
        $conf[] = $mapped;
      }
    }

    if(sizeof($conf) > 0) return implode(',', $conf);
    return false;
  }

  function TransformWebExtendedKeyUsage($fromWeb)
  {
    $keyMapping = array(
      "EKUCritical" => "critical",
      "EKUServerAuth" => "serverAuth",
      "EKUClientAuth" => "clientAuth",
      "EKUCodeSigning" => "codeSigning",
      "EKUEmailProtection" => "emailProtection",
      "EKUTimeStamping" => "timeStamping",
      "EKUOCSPSigning" => "OCSPSigning"
    );

    return ToConfString($fromWeb, $keyMapping);
  }

  function TransformWebKeyUsage($fromWeb)
  {
    $keyMapping = [
      "KUCritical" => "critical",
      "KUDigitalSignature" => "digitalSignature",
      "KUKeyEncipherment" => "keyEncipherment",
      "KUKeyAgreement" => "keyAgreement",
      "KUDataEncipherment" => "dataEncipherment",
      "KUCertSign" => "keyCertSign",
      "KUCRLSign" => "cRLSign",
      "KUEncipherOnly" => "encipherOnly",
      "KUDecipherOnly" => "decipherOnly",
      "KUNonRepudiation" => "nonRepudiation"
    ];

    return ToConfString($fromWeb, $keyMapping);
  }

  function TransformWebSanConfigs($fromWeb)
  {
    $san = str_ensure_defined($fromWeb["san"]);

    if($san == '')
    {
      return false;
    }

    $critical = str_ensure_defined($fromWeb['SanCritical']);
    
    if($critical)
    {
      $san = "critical,$san";
    }

    return $san;
  }

  function TransformWebBasicConstraints($fromWeb)
  {
    if(str_ensure_defined($fromWeb['BCInclude']) !== '1') return false;
    
    $basicConstraints = array();
    
    if(str_ensure_defined($fromWeb['BCCritical']) === '1')
    {
      $basicConstraints[] = 'critical';
    }
    
    if(str_ensure_defined($fromWeb['BCCAcertificate']) === '1')
    {
      $basicConstraints[] = 'CA:true';
      $path_len = int_ensure_defined($_POST['BCpathlen']);
      if($path_len > 0) $basicConstraints[] = "pathlen:$path_len";
    }
    else
    {
      $basicConstraints[] = 'CA:false';
    }
    
    return implode(',', $basicConstraints);
  }
  
  function TransformWebCertExtensions($fromWeb)
  {
    $extensions = [];

    $sanConfigs = TransformWebSanConfigs($fromWeb);

    if($sanConfigs)
    {
      $extensions['subjectAltName'] = $sanConfigs;
    }

    $basicConstraints = TransformWebBasicConstraints($fromWeb);

    if($basicConstraints)
    {
      $extensions["basicConstraints"] = $basicConstraints;
    }

    $keyUsage = TransformWebKeyUsage($fromWeb);

    if($keyUsage)
    {
      $extensions["keyUsage"] = $keyUsage;
    }

    $extendedKeyUsage = TransformWebExtendedKeyUsage($fromWeb);

    if($extendedKeyUsage)
    {
      $extensions["extendedKeyUsage"] = $extendedKeyUsage;
    }

    return $extensions;
  }

  function TransformWebCertConfig($fromWeb)
  {
    $configsOut = [];
    $configsOut["key"] = TransfromWebPrivKey($fromWeb);
    $configsOut["distinguishedName"] = TransfromWebDistinguishedName($fromWeb);
    $configsOut["extensions"] = TransformWebCertExtensions($fromWeb);

    return $configsOut;
  }

  function GetLetsEncryptAccount()
  {
    global $dbintf;
    $url = $dbintf->GetConfigValue("RegLetsEncryptAccount");
    $keyFile = "file://" . $dbintf->GetConfigValue("RegLetsEncryptKey");

    $account = ["key" => ImportKey($keyFile), "url" => $url];
    return $account;
  }
?>