<?php
define('SCIM',1);

define("USER_SCHEMA", "urn:ietf:params:scim:schemas:core:2.0:User");
define("GROUP_SCHEMA", "urn:ietf:params:scim:schemas:core:2.0:Group");
define("SCHEMA_SCHEMA", "urn:ietf:params:scim:schemas:core:2.0:Schema");
define("LISTRESPONSE_SCHEMA", "urn:ietf:params:scim:api:messages:2.0:ListResponse");
define("ERROR_SCHEMA", "urn:ietf:params:scim:api:messages:2.0:Error");
define("PATCH_SCHEMA", "urn:ietf:params:scim:api:messages:2.0:PatchOp");

define("USER_USERNAME_DESCRIPTION", "Unique identifier for the User, typically used by the user to directly authenticate to the service provider. Each User MUST include a non-empty userName value. This identifier MUST be unique across the service provider's entire set of Users. REQUIRED.");
define("USER_ACTIVE_DESCRIPTION", "A Boolean value indicating the User's administrative status.");
define("USER_DISPLAYNAME_DESCRIPTION", "The name of the User, suitable for display to end-users. The name SHOULD be the full name of the User being described, if known.");
define("USER_EMAILS_DESCRIPTION", "Email addresses for the user. The value SHOULD be canonicalized by the service provider, e.g., \"bjensen@example.com\" instead of \"bjensen@EXAMPLE.COM\". Canonical type values of \"work\", \"home\", and \"other\".");
define("USER_PHONES_DESCRIPTION", "Phone numbers for the User. The value SHOULD be canonicalized by the service provider according to the format specified in RFC 3966, e.g., \"tel:+1-201-555-0123\". Canonical type values of \"work\", \"home\", \"mobile\", \"fax\", \"pager\", and \"other\".");

define("GROUP_DISPLAYNAME_DESCRIPTION", "The name of the User, suitable for display to end-users.  The name SHOULD be the full name of the User being described, if known.");
define("GROUP_MEMBERS_DESCRIPTION", "A list of members of the Group.");

define ("TYPE_DESCRIPTION", "A label indicating the attribute's function");
define ("VALUE_DESCRIPTION", "The significant value for the attribute");
define ("PRIMARY_DESCRIPTION", "A Boolean value indicating the 'primary' or preferred attribute value for this attribute, The primary attribute value 'true' MUST appear no more than once.");

class ScimException extends Exception
{
  var $scimType = null;
  // Redefine the exception so message isn't optional
  public function __construct($message, $code = 0, $scimType = null, Throwable $previous = null) 
  {
    // some code
    // make sure everything is assigned properly
    parent::__construct($message, $code, $previous);
    $this->scimType = $scimType;
  }
}

function Paginate($resources, $startIndex, $total)
{
  return [
    "schemas" => [
      LISTRESPONSE_SCHEMA,
    ],
    "totalResults" => $total,
    "itemsPerPage" => count($resources),
    "startIndex" => $startIndex,
    "Resources" => $resources
  ];
}

function IsAuthorized($parentInfo)
{
  global $dbintf;

	$enabled = (int) $dbintf->GetConfigValue("EnableSCIM");

	if(!$enabled)
	{
		return false;
	}
  
  if($parentInfo['scimenabled'] < 1)
  {
    return false;
  }

  $authorization = $_SERVER['HTTP_AUTHORIZATION'] ?? null;

  if(!isset($authorization))
  {
    return false;
  }

  $authorization = $parts = preg_split('/\s+/', $authorization);

  if(count($authorization) < 2)
  {
    return false;
  }

  if(strtolower($authorization[0]) != "bearer")
  {
    return false;
  }

  $token = $parentInfo['scimtoken'];
	
	if(!isset($token) || $token == "")
	{
		return false;
	}

  if(!DBInterface::VerifyPwd($authorization[1], $token))
  {
    return false;
  }

  return true;
}

function IncludeAttribute($attribute, $attributes, $excludedAttributes)
{
	if(count($attributes) > 0)
	{
		return in_array($attribute, $attributes);
	}
	else if(count($excludedAttributes) > 0)
	{
		return !in_array($attribute, $excludedAttributes);
	}
	else
	{
		return true;
	}
}

function IsExcludedAttribute($attribute, $excludedAttributes)
{
	if(count($excludedAttributes) > 0)
	{
		return !in_array($attribute, $excludedAttributes);
	}

	return false;
}

function CreateComparison($column, $op, $value, &$toBind, $caseSensitive=true)
{
  switch($op)
  {
    case "eq":
      $toBind[] = $value;
      if($caseSensitive)
      {
        return "$column=?";
      }
      else
      {
        return "$column ILIKE ?";
      }
    case "ne":
      $toBind[] = $value;
      if($caseSensitive)
      {
        return "$column != ?";
      }
      else
      {
        return "NOT $column ILIKE ?";
      }
    case "co":
      $toBind[] = "%$value%";
      $sqlComp = $caseSensitive ? "LIKE" : "ILIKE";
      return "$column $sqlComp ?";
    case "sw":
      $toBind[] = "%$value";
      $sqlComp = $caseSensitive ? "LIKE" : "ILIKE";
      return "$column $sqlComp ?";
    case "ew":
      $toBind[] = "$value%";
      $sqlComp = $caseSensitive ? "LIKE" : "ILIKE";
      return "$column $sqlComp ?'";
    case "gt":
      $toBind[] = $value;
      return "$column > ?";
    case "ge":
      $toBind[] = $value;
      return "$column >= ?";
    case "lt":
      $toBind[] = $value;
      return "$column < ?";
    case "le":
      $toBind[] = $value;
      return "$column <= ?";
    case "pr":
      return "$column IS NOT NULL";
  }
}


function HandleConjuction($node, &$toBind, $ComparisonFunc)
{
	$factors = $node->GetFactors();
	$filter = ' TRUE';
	foreach($factors as $childNode)
	{
		$filter .= ' AND ' . HandleNode($childNode, $toBind, $ComparisonFunc);
	}

	return '(' . $filter . ')';
}

function HandleDisjuction($node, &$toBind, $ComparisonFunc)
{
	$terms = $node->GetTerms();
	$filter = ' FALSE';
	foreach($terms as $childNode)
	{
		$filter .= ' OR ' . HandleNode($childNode, $toBind, $ComparisonFunc);
	}

	return '(' . $filter . ')';
}

function HandleNegation($node, &$toBind, $ComparisonFunc)
{
  return ' NOT' . ' (' . HandleNode($node->filter, $toBind, $ComparisonFunc) . ')';
}

function HandleNode($node, &$toBind, $ComparisonFunc)
{
	if($node instanceof Tmilos\ScimFilterParser\Ast\Conjunction)
	{
    return HandleConjuction($node, $toBind, $ComparisonFunc);
	}
  else if($node instanceof Tmilos\ScimFilterParser\Ast\Disjunction)
  {
    return HandleDisjuction($node, $toBind, $ComparisonFunc);
  }
  else if($node instanceof Tmilos\ScimFilterParser\Ast\Negation)
  {
    return HandleNegation($node, $toBind, $ComparisonFunc);
  }
  else if($node instanceof Tmilos\ScimFilterParser\Ast\ComparisonExpression)
  {
    return $ComparisonFunc($node, $toBind);
  }
  else
  {
    throw new ScimException("Unsupported construct in filter", 400, "invalidFilter");
  }
}

function ToSqlFilter($node, $ComparisonFunc)
{
  $toBind = [];
	$filter = HandleNode($node, $toBind, $ComparisonFunc);

  return [
    "toBind" => $toBind,
    "filter" => $filter
  ];
}

function CheckPatchSchema($patch)
{
  $okay = false;
  foreach($patch['schemas'] as $schema)
  {
    if($schema == PATCH_SCHEMA)
    {
      $okay = true;
			break;
    }
  }

  return $okay;
}

function scimlg($msg)
{
	global $requestID;
	$parentGroup = $_GET['group'];
	lg("$requestID $parentGroup -- $msg");
}

?>