<?php

require_once "../apresa/conf/voip_config.php";
require_once "scim_general.php";
require_once "db.php";
require_once "actions.php";
require_once "datafile.php";
require_once 'signaling.php';
require_once '../apresa/vendor/autoload.php';

use Tmilos\ScimFilterParser\Parser;
use Tmilos\ScimFilterParser\Mode;

function CheckSchema($newGroup)
{
  $okay = false;
  foreach($newGroup['schemas'] as $schema)
  {
    if($schema == GROUP_SCHEMA)
    {
      $okay = true;
			break;
    }
  }

  return $okay;
}

function AddMembersToGroup($group, $members)
{
	global $dbintf;

	$memberlist= [];
	foreach($members as $member)
	{
		$member = array_change_key_case($member, CASE_LOWER);

		if(!isset($member["value"])) continue;
		$user = $dbintf->GetScimUser((int) $member["value"], $group['scimparent']);
		if($user === false) continue;
		$memberlist[] = $user['accid'];
	}

	if(count($memberlist) > 0)
	{
		if(!$dbintf->SetGroupMembers($group['accid'], $memberlist))
		{
			throw new ScimException($dbintf->lasterr, 500);
		}
	}
}

function NewGroup($newGroup, $parentInfo)
{
	global $dbintf;

	if(!isset($newGroup))
	{
		throw new ScimException("No request body available", 400, "invalidSyntax");
	}

  $newGroup = array_change_key_case($newGroup, CASE_LOWER);

  if(!CheckSchema($newGroup))
  {
    throw new ScimException("Unsupported schema", 400, "invalidSyntax");
  }

  if(isset($_GET['id']))
  {
    throw new ScimException("No group id should be specified on creation of a new one", 400, "invalidValue");
  }

  if(!isset($newGroup['displayname']))
  {
    throw new ScimException("Group displayname is required", 400, "invalidValue");
  }

  if($dbintf->GroupExists($newGroup['displayname']))
  {
    throw new ScimException("A group with the displayname exists already", 409);
  }

	$dbintf->StartTransaction();

	if(!$dbintf->AddGroup($id, [
		"name" => $newGroup['displayname'],
		"typegrp" => 0,
		"super" => $parentInfo['accid'],
		"idp" => $parentInfo['idp'],
    "scimenabled" => -1
	]))
	{
		throw new ScimException($dbintf->lasterr, 500);
	}

	$groupInfo = $dbintf->GetGroupInfo($id);
	$groupInfo["scimparent"] = $parentInfo["id"];

	$time = new DateTime("now", new DateTimeZone("UTC"));

  $scimData = [
    "created" => $time->format(DateTimeInterface::ATOM),
    "updated" => $time->format(DateTimeInterface::ATOM),
    "accid" => $groupInfo['accid'],
    "externalid" => $newGroup['externalid'],
		"scimparent" => $parentInfo["id"]
  ];

	$dbintf->InsertScimData($scimData);

	if(isset($newGroup["members"]))
	{
		AddMembersToGroup($groupInfo, $newGroup['members']);
	}

	$dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
	return GetGroup($id, $parentInfo);
}

function GetGroup($id, $parentInfo)
{
	global $dbintf;

  if(!isset($id))
  {
    throw new ScimException("No group id specified", 400, "invalidValue");
  }

	$group = $dbintf->GetScimGroup((int) $id, $parentInfo['id']);

	if($group === false)
  {
    throw new ScimException("Group for $id not found", 404);
  }

	return DbToScimGroup($group);
}

function DbToScimGroup($group)
{
	global $dbintf;

	$attributes = isset($_GET['attributes']) ? explode(',', strtolower($_GET['attributes'])) : [];
	$excludedAttributes = isset($_GET['excludedAttributes']) ? explode(',', strtolower($_GET['excludedAttributes'])) : [];

	$UseAttribute = function($attribute) use ($attributes, $excludedAttributes)
	{
		return IncludeAttribute($attribute, $attributes, $excludedAttributes);
	};

	$created = new DateTime($group['created']);
  $created->setTimezone(new DateTimeZone("UTC"));
  $updated = new DateTime($group['updated']);
  $updated->setTimezone(new DateTimeZone("UTC"));

  $result = [
		"externalId" => str_ensure_defined($group['externalid']),
		"id" => (string) $group['id'],
		"schemas" => [GROUP_SCHEMA],
		"meta" => [
      "resourceType" => "Group",
			"created" => $created->format(DateTimeInterface::ATOM),
			"lastModified" => $created->format(DateTimeInterface::ATOM)
		]
	];

	if($UseAttribute('displayname'))
	{
		$result["displayName"] = $group["name"];
	}

	if($UseAttribute('members'))
	{
		$scimMembers = [];
		$members = $dbintf->GetGroupMembers($group['accid']);
		foreach($members as $member)
		{
			$memberid = $dbintf->GetScimUserIdFromAccid($member);
			if($memberid === false) continue;

			$scimMembers[] = [
				"value" => (string) $memberid
			];
		}

		$result["members"] = $scimMembers;
	}

	return $result;
}

function DeleteScimGroup($id, $parentInfo)
{
  global $dbintf;

  if(!isset($id))
  {
    throw new ScimException("No group id specified", 404);
  }

  $dbintf->StartTransaction();

  // Only allow deletion of groups that were also created through SCIM
  $group = $dbintf->GetScimGroup((int) $id, $parentInfo["id"]);

  if($group === false)
  {
    throw new ScimException("Group not found", 404);
  }

  if(!$dbintf->DeleteGroup($id))
  {
    throw new ScimException($dbintf->lasterr, 500);
  }

  $dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
}

function ReplaceGroup($id, $newGroup, $parentInfo)
{
	global $dbintf;

	if(!isset($id))
	{
		throw new ScimException("No group id specified", 400, "invalidValue");
	}

	if(!isset($newGroup))
	{
		throw new ScimException("No request body available", 400, "invalidSyntax");
	}

  $newGroup = array_change_key_case($newGroup, CASE_LOWER);

	if(!isset($newGroup['displayname']))
	{
		throw new ScimException("displayName has not been set", 400, "invalidValue");
	}

  if(!CheckSchema($newGroup))
  {
    throw new ScimException("Unsupported schema", 400, "invalidSyntax");
  }

	$dbintf->StartTransaction();

	$group = $dbintf->GetScimGroup((int) $id, $parentInfo["id"]);
  if($group === false)
  {
    throw new ScimException("Group not found", 404);
  }

  if($newGroup["displayname"] != $group["name"] && $dbintf->GroupExists($newGroup['displayname']))
  {
    throw new ScimException("A group with the displayname exists already", 409);
  }

	if(!$dbintf->EditGroup($id, ["name" => $newGroup["displayname"]]))
	{
		throw new ScimException($dbintf->lasterr, 500);
	}

	if(isset($newGroup["members"]))
	{
		AddMembersToGroup($group, $newGroup["members"]);
	}

	$time = new DateTime("now", new DateTimeZone("UTC"));

	$scimData =
	[
		"updated" => $time->format(DateTimeInterface::ATOM),
		"externalid" => str_ensure_defined($newGroup['externalid'])
	];

	if(!$dbintf->EditScimData($group['accid'], $scimData))
	{
		throw new ScimException("Could not write updated group information to database", 500);
	}

	$dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
}

function GroupSqlComparison($node, &$toBind)
{
  $attributeName =  strtolower(implode(".", $node->attributePath->attributeNames));
  $validOps = [];

  if($attributeName == "displayname")
  {
    $validOps = ["eq", "ne", "co", "sw", "ew"];
    if(!in_array($node->operator, $validOps))
    {
      throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
    }
    return CreateComparison("g.name", $node->operator, $node->compareValue, $toBind);
  }
	else 
	{
		throw new ScimException("Unsupported filter attribute '$attributeName'", 400, "invalidFilter");
	}
}

function GetGroups($parentInfo)
{
	global $dbintf;

  $filter = null;
	if(isset($_GET['filter']))
	{
		$parser = new Parser(Mode::FILTER());
		$node = $parser->parse($_GET['filter']);
		$filter = ToSqlFilter($node, "GroupSqlComparison");
	}

  $total = $dbintf->GetScimTotalGroups($parentInfo['id'], $filter);

  if($total === false)
  {
    throw new ScimException("Database error when retrieving matching number of groups", 500);
  }

  $startIndex = 1;
  if(isset($_GET['startIndex']))
  {
    $startIndex = intval($_GET['startIndex']);

    if($startIndex > 0)
    {
      $filter['startIndex'] = $startIndex-1;
    }
    else
    {
      $startIndex = 1;
    }
  }

  if(isset($_GET['count']))
  {
    $filter['count'] = intval($_GET['count']);
  }
  else if($total > 100)
  {
    $filter['count'] = 100;
  }

  if(isset($filter['startIndex']) && !isset($filter['count']))
  {
    $filter['count'] = $total;
  }

	$groups = $dbintf->GetScimGroups($parentInfo['id'], $filter);

  if($groups === false)
  {
    throw new ScimException("Database error when retrieving groups", 500);
  }
	
  $scimGroups = [];
	
	foreach($groups as $group)
	{
		$scimGroups[] = DbToScimGroup($group);
	}

	return Paginate($scimGroups, $startIndex, $total);
}

function UpdateGroupName(&$group, $value, $op)
{
  global $dbintf;
  
  if($op == "remove")
  {
    throw new ScimException("Cannot remove group name", 400, "invalidValue");
  }

  if(!isset($value))
  {
    throw new ScimException("Cannot set group name to empty", 400, "invalidValue");
  }

  if($value != $group['name'] && $dbintf->GroupExists($value))
  {
    throw new ScimException("A user with the username exists already", 409);
  }

  if($group["name"] != $value)
  {
    $group["updated"] = true;
    $group["name"] = $value;
  }
}

function UpdateGroupMembers(&$group, $value, $op)
{
  global $dbintf;

  $existingMembers = $dbintf->GetGroupMembers($group['accid']);
  if($op == "remove")
  {
    if(!isset($value))
    {
      $dbintf->SetGroupMembers($group['accid'], []);
      $group["updated"] = true;
    }
    else
    {
      if(!is_array($value))
      {
        throw new ScimException("Expected value to be an array", 400, "invalidValue");
      }
      
      $toRemove = array_map(function($member) use($dbintf, $group){
        $user = $dbintf->GetScimUser((int) $member["value"], $group["scimparent"]);
        if($user === false)
        {
          throw new ScimException("Cannot find user " . $member["value"], 400, "invalidValue");
        }
        return $user["accid"];
      }, $value);

      $newMembers = [];

      foreach($existingMembers as $existing)
      {
        if(!in_array($existing, $toRemove))
        {
          $newMembers[] = $existing;
        }
      }

      if($existingMembers != $newMembers)
      {
        $group["updated"] = true;
      }

      if(!$dbintf->SetGroupMembers($group['accid'], $newMembers))
      {
        throw new ScimException($dbintf->lasterr, 500);
      }
    }
  }
  else if($op == "replace")
  {
    if(!is_array($value))
    {
      throw new ScimException("Expected value to be an array", 400, "invalidValue");
    }

    $members = [];
    foreach($value as $newMember)
    {
      $user = $dbintf->GetScimUser((int) $newMember["value"], $group['scimparent']);
      if($user === false)
      {
        throw new ScimException("Cannot find user " . $newMember["value"], 400, "invalidValue");
      }

      $members[] = $user['accid'];
    }
    
    $members =  array_unique($members);
    if($members != $existingMembers)
    {
      $group["updated"] = true;
      if(!$dbintf->SetGroupMembers($group['accid'],$members))
      {
        throw new ScimException($dbintf->lasterr, 500);
      }
    }
  }
  else if($op == "add")
  {
    $members = $dbintf->GetGroupMembers($group['accid']);
    foreach($value as $newMember)
    {
      $user = $dbintf->GetScimUser((int) $newMember["value"], $group['scimparent']);
      if($user === false)
      {
        throw new ScimException("Cannot find user " . $newMember["value"], 400, "invalidValue");
      }

      $members[] = $user['accid'];
    }

    $members = array_unique($members);
    if($members != $existingMembers)
    {
      $group["updated"] = true;
      if(!$dbintf->SetGroupMembers($group['accid'], $members))
      {
        throw new ScimException($dbintf->lasterr, 500);
      }
    }
  }
}

function MemberSqlComparison($node, &$toBind)
{
  $attributeName =  strtolower(implode(".", $node->attributePath->attributeNames));
  switch($attributeName)
  {
    case "value": 
      if($node->operator != "eq" && $node->operator != "ne")
      {
        throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
      }

      return CreateComparison("u.id", $node->operator, $node->compareValue, $toBind);
      break;
    default:
      throw new ScimException("Unsupported filter attribute '$attributeName'", 400, "invalidFilter");
  }
}

function UpdateGroupMembersWithFilter($patch, $filter, $attributePath, &$group)
{
  global $dbintf;
  if(isset($attributePath))
  {
    throw new ScimException("Patching subattribute of members not supported", 400, "invalidFilter");
  }

  $filter = ToSqlFilter($filter, "MemberSqlComparison");
  $selectedMembers = $dbintf->GetScimGroupMembers($group['accid'], $filter);
  
  $op = strtolower($patch["op"]);

  if($op == "remove" || $op == "replace" || $op == "add")
  {
    if($op == "remove" && isset($patch["value"]))
    {
      throw new ScimException("Value unexpectedly been set on a patch operation to remove members specified with a filter", 400, "invalidFilter");
    }

    if(($op == "replace" || $op == "add") && !is_array($patch["value"]))
    {
      throw new ScimException("Expected value to be an array", 400, "invalidValue");
    }

    if($op == "replace" && count($selectedMembers) == 0)
    {
      throw new ScimException("No records have been selected for replace operation", 400, "invalidPath");
    }

    $toRemove = array_map(function($member){
      return $member["accid"];
    }, $selectedMembers);

    $existingMembers = $dbintf->GetGroupMembers($group['accid']);
    $newMembers = [];

    foreach($existingMembers as $existing)
    {
      if(!in_array($existing, $toRemove))
      {
        $newMembers[] = $existing;
      }
    }

    if($op == "replace" || $op == "add")
    {
      $value = $patch["value"];
			if(isset($value["value"]))
			{
				$value = [$value];
			}

      foreach($value as $newMember)
      {
        $user = $dbintf->GetScimUser((int) $newMember["value"], $group['scimparent']);
        if($user === false)
        {
          throw new ScimException("Cannot find user " . $newMember["value"], 400, "invalidValue");
        }

        $newMembers[] = $user['accid'];
      }
    }

    $newMembers = array_unique($newMembers);

    if($newMembers != $existingMembers)
    {
      $group["updated"] = true;
      if(!$dbintf->SetGroupMembers($group['accid'], array_unique($newMembers)))
      {
        throw new ScimException($dbintf->lasterr, 500);
      }
    }
  }
}

function PatchAttributePath($patch, $attributePath, &$group)
{
  $value = $patch["value"] ?? null;
  $op = strtolower($patch["op"]);

  switch(strtolower($attributePath->attributeNames[0]))
  {
    case 'displayname':
      UpdateGroupName($group, $value, $op);
      break;  
    case 'members':
      UpdateGroupMembers($group, $value, $op);
      break;
    default:
      throw new ScimException("Unsupported patch path: " . implode(".", $attributePath->attributeNames), 400, "invalidPath");
  }
}

function PatchValuePath($patch, $valuePath, $attributePath, &$group)
{
  $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\ValuePath::class, 'attributePath');
  $reflectionProperty->setAccessible(true);
  $innerAttributePath = $reflectionProperty->getValue($valuePath);

  $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\ValuePath::class, 'filter');
  $reflectionProperty->setAccessible(true);
  $filter = $reflectionProperty->getValue($valuePath);

  switch(strtolower($innerAttributePath->attributeNames[0]))
  {
    case 'members':
      UpdateGroupMembersWithFilter($patch, $filter, $attributePath, $group);
      break;
    default:
      throw new ScimException("Unsupported patch path: " . implode(".", $innerAttributePath->attributeNames), 400, "invalidPath");
  }
}

function PatchWithoutPath($value, &$group)
{
  global $dbintf;
  if(!isset($value))
  {
    throw new ScimException("Expected a value in the patch operation", 400, "invalidValue");
  }

  $value = array_change_key_case($value, CASE_LOWER);

  if(isset($value["displayname"]))
  {
    if($value['displayname'] != $group['name'] && $dbintf->GroupExists($value['displayname']))
    {
      throw new ScimException("A group with the displayname exists already", 409);
    }

    $group['name'] = $value["displayname"];
  }

  if(isset($value["externalid"]))
  {
    $group["externalid"] = $value["externalid"];
  }
}

function DoPatchOperation($patch, &$group)
{
  $patch = array_change_key_case($patch, CASE_LOWER);

  if(isset($patch['path']))
  {
    $parser = new Parser(Mode::PATH());
		$node = $parser->parse($patch['path']);

    // Hack to get at properties that library sets to private...

    $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\Path::class, 'attributePath');
    $reflectionProperty->setAccessible(true);
    $attributePath = $reflectionProperty->getValue($node);

    $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\Path::class, 'valuePath');
    $reflectionProperty->setAccessible(true);
    $valuePath = $reflectionProperty->getValue($node);

    if(!isset($valuePath))
    {
      PatchAttributePath($patch, $attributePath, $group);
    }
    else
    {
      PatchValuePath($patch, $valuePath, $attributePath, $group);
    }
  }
  else
  {
    if(strtolower($patch["op"]) == "remove")
    {
      throw new ScimException("No target path has been specified for remove operation", 400, "noTarget");
    }
    else
    {
      PatchWithoutPath($patch["value"], $group);
    }
  }
}

function PatchGroup($id, $patch, $parentInfo)
{
  global $dbintf;

  if(!isset($id))
	{
		throw new ScimException("No group id specified", 400, "invalidValue");
	}

	if(!isset($patch))
	{
		throw new ScimException("No request body available", 400, "invalidSyntax");
	}

  $patch = array_change_key_case($patch, CASE_LOWER);

  if(!CheckPatchSchema($patch))
  {
    throw new ScimException("Unsupported schema", 400, "invalidSyntax");
  }

  $group = $dbintf->GetScimGroup((int) $id, $parentInfo["id"]);

  if($group === false)
  {
    throw new ScimException("Group not found", 404);
  }

  $dbintf->StartTransaction();

  foreach($patch["operations"] as $operation)
  {
    DoPatchOperation($operation, $group);
  }

  if($group["updated"])
  {

    $dbgroup = [
      "name" => $group["name"]
    ];

    $time = new DateTime("now", new DateTimeZone("UTC"));

    $scimData =
    [
      "updated" => $time->format(DateTimeInterface::ATOM),
      "externalid" => $group["externalid"]
    ];

    if(!$dbintf->EditGroup($id, $dbgroup))
    {
      throw new ScimException($dbintf->lasterr, 500);
    }

    if(!$dbintf->EditScimData($group['accid'], $scimData))
    {
      throw new ScimException("Could not write updated group information to database", 500);
    }
  }

  $dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
}

function QueryParamLog()
{
  $logs = [];
  if(isset($_GET['filter']))
  {
    $logs[] = 'filter: ' . $_GET['filter'];
  }

  if(isset($_GET['attributes']))
  {
    $logs[] = 'attributes: ' . $_GET['attributes'];
  }

  if(isset($_GET['excludedAttributes']))
  {
    $logs[] = 'excludedAttributes: ' . $_GET['excludedAttributes'];
  }

  return 'QueryParams: ' . implode(",", $logs);
}


function HandleRequest()
{
	if(!isset($_GET['group']))
	{
		throw new ScimException("Not found", 404);
	}

	global $requestID;
	$requestID = uniqid();
	$parentGroup = $_GET['group'];

  global $dbintf;
  InitDB();

	$parentInfo = $dbintf->GetSingleGroupInfoByName($parentGroup);
	if($parentInfo === false)
	{
		throw new ScimException("Not found", 404);
	}

  if(!IsAuthorized($parentInfo))
  {
    throw new ScimException("Unauthorized", 401);
  }

	$dbintf->SetAdmin();

  $method = $_SERVER['REQUEST_METHOD'];

  switch($_SERVER['REQUEST_METHOD'])
  {   
    case 'GET':
			if(isset($_GET['id']))
			{
        scimlg("SCIM group get with ID: " . $_GET['id']  . ", " . QueryParamLog());
      	$result = GetGroup($_GET['id'], $parentInfo);
      }
      else
      {
        scimlg("SCIM group get without ID"  . ", " . QueryParamLog());
        $result = GetGroups($parentInfo);
      }
      
      $response = json_encode($result);
      scimlg("SCIM Group GET response: " . $response);
      header('Content-Type: application/scim+json');
      echo $response;
      break;
    case 'POST':
      $input = file_get_contents("php://input");
      $body = json_decode($input, true);
      scimlg("SCIM Group POST request: " . json_encode($body)  . ", " . QueryParamLog());
      $result = NewGroup($body, $parentInfo);
      $response = json_encode($result);
      scimlg("SCIM Group POST response: " . $response);
      http_response_code(201);
      header('Content-Type: application/scim+json');
      echo $response;
      break;
    case 'PATCH':
      $input = file_get_contents("php://input");
      $body = json_decode($input, true);
      scimlg("SCIM Group " .  $_GET['id'] . " PATCH request: " . json_encode($body)  . ", " . QueryParamLog());
      PatchGroup($_GET['id'], $body, $parentInfo);
      http_response_code(200);
      break;
		case 'PUT':
			$input = file_get_contents("php://input");
			$body = json_decode($input, true);
      scimlg("SCIM Group " . $_GET['id'] . " PUT request: " . json_encode($body)  . ", " . QueryParamLog());
			ReplaceGroup($_GET['id'], $body, $parentInfo);
			http_response_code(200);
			break;
    case 'DELETE':
      scimlg("SCIM Group " . $_GET['id'] . " DELETE request" . ", " . QueryParamLog());
      DeleteScimGroup($_GET['id'], $parentInfo);
      http_response_code(204);
      break;
    default:
      throw new ScimException("This method is not supported: $method", 501);
  }
}

function CancelDB()
{
  global $dbintf;
  if(isset($dbintf))
  {
    while($dbintf->transactionLevel > 0)
    {
      $dbintf->CancelTransaction();
    }
  }
}

try
{
  HandleRequest();
}
catch(ScimException $exception)
{
  $result = [
    "schemas" => [ERROR_SCHEMA],
    "detail" => $exception->getMessage(),
    "status" => $exception->getCode()
  ];

  if(isset($exception->scimType))
  {
    $result["scimType"] = $exception->scimType;
  }

  $code = $exception->getCode();

  $response = json_encode($result);
  scimlg('Scim exception on group: ' . $response);

  header('Content-Type: application/scim+json');
  http_response_code($code);
  echo $response;
}
catch(Tmilos\ScimFilterParser\Error\FilterException | Tmilos\Lexer\Error\UnknownTokenException $exception)
{
  $result = [
    "schemas" => [ERROR_SCHEMA],
    "detail" => "Filter syntax error",
    "scimType" => "invalidFilter",
    "status" => 400
  ];

  $response = json_encode($result);
  scimlg('Scim filter parse exception on user: ' . $response);

  header('Content-Type: application/scim+json');
  http_response_code(400);
  echo $response;
}
catch(Exception $exception)
{
	$result = [
    "schemas" => [ERROR_SCHEMA],
    "detail" => "An internal error occured",
    "status" => 500
  ];

	$response = json_encode($result);
  scimlg('Exception on user: ' . $exception->getMessage());
	header('Content-Type: application/scim+json');
  http_response_code(500);
  echo $response;
}
finally
{
  CancelDB();
}

?>