<?php

require_once "../apresa/conf/voip_config.php";
require_once "scim_general.php";
require_once "db.php";
require_once "actions.php";
require_once "datafile.php";
require_once 'signaling.php';
require_once '../apresa/vendor/autoload.php';

use Tmilos\ScimFilterParser\Parser;
use Tmilos\ScimFilterParser\Mode;

function CheckSchema($newUser)
{
  $okay = false;
  foreach($newUser['schemas'] as $schema)
  {
    if($schema == USER_SCHEMA)
    {
      $okay = true;
			break;
    }
  }

  return $okay;
}

function ToBool(&$value)
{
	if(!isset($value)) return false;

  $trueVals = [true, 1, "1", "true", "yes", "y"];
  $falseVals = [false, 0, "0", "false", "no", "n"];

  if(is_string($value))
  {
    $value = strtolower($value);
  }

  if(in_array($value, $trueVals, true)) return true;

  if(in_array($value, $falseVals, false)) return false;

  throw new ScimException("Cannot convert $value to boolean", 400, "invalidValue");
}

function GetUser($id, $parentInfo)
{
  global $dbintf;

  if(!isset($id))
  {
    throw new ScimException("No user id specified", 400, "invalidValue");
  }

  $user = $dbintf->GetScimUser($id, $parentInfo['id']);

  if($user === false)
  {
    throw new ScimException("User for $id not found", 404);
  }

  $result = DbToScimUser($user);

  return $result;
}

function GetEmailFromScim(&$dbuser, $scimuser)
{
	$email = null;
	if(isset($scimuser['emails']))
	{
		foreach($scimuser['emails'] as $newEmail)
		{
			$newEmail = array_change_key_case($newEmail, CASE_LOWER);
			if($email == null || ToBool($newEmail['primary']))
			{
				$email = $newEmail;
			}
		}
	}

  if($email)
  {
    $dbuser['email'] = str_ensure_defined($email['value']);
    $dbuser['extra'] = ArrayToString([
      'email_type' => str_ensure_defined($email["type"]), 
      'email_primary' => ToBool($email['primary'])
    ]);
  }
}

function CreateUserPhones($newUser)
{
	global $dbintf;

	$phones = [];
	$AnyPrimary = false;

	if(isset($newUser['phonenumbers']))
	{
		foreach ($newUser['phonenumbers'] as $phonenumber)
		{
			$phonenumber = array_change_key_case($phonenumber, CASE_LOWER);

			$res = $dbintf->GetOrCreateCallerID($phonenumber['value'], '' , 1, $newUser['displayname']);

			$contactInfo = $dbintf->GetContactInfo($res['contactid']);

			if($contactInfo['name'] == '')
			{
        $displayName = $newUser['displayname'];
        $dbintf->EditContact($res['contactid'], $displayName);
			}

			$primary = ToBool($phonenumber["primary"]);

			if($primary && $AnyPrimary)
			{
				throw new ScimException("Multiple primary phone numbers", 400, "invalidValue");
			}
			else if($primary)
			{
				$AnyPrimary = true;
			}

			$phones[] = [
				"contactid" => (int) $res["contactid"],
        "scim_data" => [
          "type" => str_ensure_defined($phonenumber["type"]),
          "is_primary" => $primary
        ]
			];    
		}
	}

	return $phones;
}

function NewUser($newUser, $parentInfo)
{
  global $dbintf;

	if(!isset($newUser))
	{
		throw new ScimException("No request body available", 400, "invalidSyntax");
	}

  $newUser = array_change_key_case($newUser, CASE_LOWER);

  if(!CheckSchema($newUser))
  {
    throw new ScimException("Unsupported schema", 400, "invalidSyntax");
  }

  if(isset($_GET['id']))
  {
    throw new ScimException("No user id should be specified on creation of a new one", 400, "invalidValue");
  }

  if(!isset($newUser['username']))
  {
    throw new ScimException("User name is required", 400, "invalidValue");
  }

  if($dbintf->UsrExists($newUser['username']))
  {
    throw new ScimException("A user with the username exists already", 409);
  }

  $dbintf->StartTransaction();

  $extra = [];

  $dbUser = 
  [
    "usr" => $newUser['username'],
    "enabled" => ToBool($newUser['active']) ? 1:0,
    "name" => str_ensure_defined($newUser['displayname']),
		"logonmethod" => 2,
    "rod" => 0
  ];

	GetEmailFromScim($dbUser, $newUser);

  if(!$dbintf->AddUserAccount($id, $dbUser))
  {
    throw new ScimException($dbintf->lasterr, 500);
  }

  $time = new DateTime("now", new DateTimeZone("UTC"));

  $createdUser = $dbintf->GetUserInfo($id);

  $scimData = [
    "created" => $time->format(DateTimeInterface::ATOM),
    "updated" => $time->format(DateTimeInterface::ATOM),
    "accid" => $createdUser['accid'],
    "externalid" => $newUser['externalid'] ?? null,
		"scimparent" => $parentInfo['id']
  ];

  $dbintf->InsertScimData($scimData);

	if(!$dbintf->SetGroupsOfWhichMember($createdUser['accid'], [$parentInfo['accid']]))
	{
		throw new ScimException($dbintf->lasterr, 500);
	}

	$phones = CreateUserPhones($newUser);

  if(count($phones) > 0 && !$dbintf->SetScimUserPhones($createdUser['accid'], $phones))
  {
    throw new ScimException($dbintf->lasterr, 500);
  }

  $dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
  return GetUser($id, $parentInfo);
}

function DeleteUser($id, $parentInfo)
{
  global $dbintf;

  if(!isset($id))
  {
    throw new ScimException("No user id specified", 404);
  }

  $dbintf->StartTransaction();

  // Only allow deletion of user that were also created through SCIM
  $user = $dbintf->GetScimUser($id, $parentInfo['id']);
  if($user === false)
  {
    throw new ScimException("User not found", 404);
  }

  if(!$dbintf->DeleteUserAccount($id))
  {
    throw new ScimException($dbintf->lasterr, 500);
  }
  $dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
}

function ReplaceUser($id, $newUser, $parentInfo)
{
	global $dbintf;

	if(!isset($id))
	{
		throw new ScimException("No user id specified", 400, "invalidValue");
	}

	if(!isset($newUser))
	{
		throw new ScimException("No request body available", 400, "invalidSyntax");
	}

  $newUser = array_change_key_case($newUser, CASE_LOWER);

	if(!isset($newUser['username']))
	{
		throw new ScimException("Username has not been set", 400, "invalidValue");
	}

  if(!CheckSchema($newUser))
  {
    throw new ScimException("Unsupported schema", 400, "invalidSyntax");
  }

	$dbintf->StartTransaction();

	$user = $dbintf->GetScimUser($id, $parentInfo['id']);
  if($user === false)
  {
    throw new ScimException("User not found", 404);
  }

	if($newUser['username'] != $user['usr'] && $dbintf->UsrExists($newUser['username']))
	{
		throw new ScimException("A user with the username exists already", 409);
	}

	$dbUser = [
		"usr" => $newUser['username'],
    "enabled" => ToBool($newUser['active']) ? 1:0,
    "name" => str_ensure_defined($newUser['displayname']),
	];

	$dbUser['email'] = '';
	$dbUser['extra'] = '';

	GetEmailFromScim($dbUser, $newUser);
	if(!$dbintf->EditUserAccount($id, $dbUser))
	{
		throw new ScimException($dbintf->lasterr, 500);
	}

	$time = new DateTime("now", new DateTimeZone("UTC"));

	$scimData =
	[
		"updated" => $time->format(DateTimeInterface::ATOM),
		"externalid" => str_ensure_defined($newUser['externalid'])
	];

	if(!$dbintf->EditScimData($user['accid'], $scimData))
	{
		throw new ScimException("Could not write updated user information to database", 500);
	}

	$phones = CreateUserPhones($newUser);

	if(!$dbintf->SetScimUserPhones($user['accid'], $phones))
  {
    throw new ScimException($dbintf->lasterr, 500);
  }

	$dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
}

function DbToScimUser($user)
{
	global $dbintf;

	$attributes = isset($_GET['attributes']) ? explode(',', strtolower($_GET['attributes'])) : [];
	$excludedAttributes = isset($_GET['excludedAttributes']) ? explode(',', strtolower($_GET['excludedAttributes'])) : [];

	$UseAttribute = function($attribute) use ($attributes, $excludedAttributes)
	{
		return IncludeAttribute($attribute, $attributes, $excludedAttributes);
	};

	$IsExcluded = function($attribute) use ($excludedAttributes)
	{
		return IsExcludedAttribute($attribute, $excludedAttributes);
	};

	$created = new DateTime($user['created']);
  $created->setTimezone(new DateTimeZone("UTC"));
  $updated = new DateTime($user['updated']);
  $updated->setTimezone(new DateTimeZone("UTC"));

  $result = [
		"externalId" => str_ensure_defined($user['externalid']),
		"id" => (string) $user['id'],
		"schemas" => [USER_SCHEMA],
		"meta" => [
      "resourceType" => "User",
			"created" => $created->format(DateTimeInterface::ATOM),
			"lastModified" => $created->format(DateTimeInterface::ATOM)
		]
	];

	if($UseAttribute("username"))
	{
		$result["userName"] = $user['usr'];
	}

	if($UseAttribute("active"))
	{
		$result["active"] = $user['enabled'] == "1";
	}

	if($UseAttribute("displayname") && !empty($user['name']))
	{
    $result["displayName"] = $user['name'];
	}

	if($UseAttribute("emails") && !empty($user['email']))
	{
		$extra = StringToArray($user['extra']);

		$email = [];
		if(!$IsExcluded("email.primary"))
		{
			$email["primary"] =  $extra['email_primary'] == '0' ? false : true;
		}

		if(!$IsExcluded("email.type"))
		{
			$email["type"] = empty($extra['email_type']) ? "other" : $extra["email_type"];
		}

		if(!$IsExcluded("email.value"))
		{
			$email["value"] = $user["email"];
		}

		$result["emails"] = [$email];
  }

	if($UseAttribute("phonenumbers"))
	{

		$phones = $dbintf->GetScimUserPhones($user['accid']);
    $phonelist = [];

		foreach($phones as $dbPhone)
		{
			$phone = [];

			if(!$IsExcluded("value"))
			{
				$phone["value"] = $dbPhone['callerid'];
			}

			if(!$IsExcluded("type"))
			{
				$phone["type"] = $dbPhone['scim_data']['type'];
			}

			if(!$IsExcluded("primary"))
			{
        $phone["primary"] = $dbPhone['scim_data']['is_primary'];
			}

			$phonelist[] = $phone;
		}

		if(count($phonelist) > 0)
		{
			$result["phoneNumbers"] = $phonelist;
		}
	}

  return $result;
}

function UserSqlComparison($node, &$toBind)
{
  $attributeName =  strtolower(implode(".", $node->attributePath->attributeNames));
  $validOps = [];

  if($attributeName == "username")
  {
    $validOps = ["eq", "ne", "co", "sw", "ew"];
    if(!in_array($node->operator, $validOps))
    {
      throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
    }
    return CreateComparison("u.usr", $node->operator, $node->compareValue, $toBind, false);
  }
  if($attributeName == "externalid")
  {
    $validOps = ["eq", "ne", "co", "sw", "ew"];
    if(!in_array($node->operator, $validOps))
    {
      throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
    }
    return CreateComparison("s.externalid", $node->operator, $node->compareValue, $toBind);
  }
	else 
	{
		throw new ScimException("Unsupported filter attribute '$attributeName'", 400, "invalidFilter");
	}
}

function GetUsers($parentInfo)
{
	global $dbintf;

  $filter = null;
	if(isset($_GET['filter']))
	{
		$parser = new Parser(Mode::FILTER());
		$node = $parser->parse($_GET['filter']);
		$filter = ToSqlFilter($node, "UserSqlComparison");
	}

  $total = $dbintf->GetScimTotalUsers($parentInfo['id'], $filter);

  if($total === false)
  {
    throw new ScimException("Database error when retrieving matching number of users", 500);
  }

  $startIndex = 1;
  if(isset($_GET['startIndex']))
  {
    $startIndex = intval($_GET['startIndex']);

    if($startIndex > 0)
    {
      $filter['startIndex'] = $startIndex-1;
    }
    else
    {
      $startIndex = 1;
    }
  }

  if(isset($_GET['count']))
  {
    $filter['count'] = intval($_GET['count']);
  }
  else if($total > 100)
  {
    $filter['count'] = 100;
  }

  if(isset($filter['startIndex']) && !isset($filter['count']))
  {
    $filter['count'] = $total;
  }

	$users = $dbintf->GetScimUsers($parentInfo['id'], $filter);

  if($users === false)
  {
    throw new ScimException("Database error when retrieving users", 500);
  }
	
  $scimUsers = [];
	
	foreach($users as $user)
	{
		$scimUsers[] = DbToScimUser($user);
	}

	return Paginate($scimUsers, $startIndex, $total);
}

function UpdateUserName(&$user, $value, $operation)
{
  global $dbintf;

  $operation = strtolower($operation);
  if($operation == "remove")
  {
    throw new ScimException("Cannot remove username", 400, "invalidValue");
  }

  if(!isset($value))
  {
    throw new ScimException("Cannot set username to empty", 400, "invalidValue");
  }

  if($value != $user['usr'] && $dbintf->UsrExists($value))
  {
    throw new ScimException("A user with the username exists already", 409);
  }

  $user['usr'] = $value;
}

function UpdateActive(&$user, $value, $operation)
{
  $operation = strtolower($operation);
  if($operation == "remove")
  {
    throw new ScimException("Cannot remove active", 400, "invalidValue");
  }

  if(!isset($value))
  {
    throw new ScimException("Cannot set active to empty", 400, "invalidValue");
  }

  $user['enabled'] = ToBool($value) ? 1:0;
}

function UpdateExternalID(&$user, $value, $operation)
{
  $operation = strtolower($operation);
  if($operation == "remove")
  {
    $user['externalid'] = null;
  }

  $user['externalud'] = $value;
}

function UpdateDisplayName(&$user, $value, $operation)
{
  $operation = strtolower($operation);
  if($operation == "remove")
  {
    $user['name'] = null;
  }

  $user['name'] = $value;
}

function UpdateEmails(&$user, $value, $operation)
{
  if(($operation == "add" || $operation == "replace")  && !is_array($value))
  {
    throw new ScimException("Expected a multivalued input for attribute emails", 400, "invalidValue");
  }

  if($operation == "remove")
  {
    $user['email'] = '';
    $extra = StringToArray($user['extra']);
    unset($extra['email_type']);
    unset($extra['email_primary']);
    $user['extra'] = ArrayToString($extra);
  }
  else if(($operation == "add" && empty($user['email'])) || $operation == "replace")
  {
    GetEmailFromScim($user, ["emails" => $value]);
  }
}

function UpdatePhoneNumbers(&$user, $value, $operation)
{
  global $dbintf;

  if(($operation == "add" || $operation == "replace")  && !is_array($value))
  {
    throw new ScimException("Expected a multivalued input for attribute phonenumbers", 400, "invalidValue");
  }

  if($operation == "remove")
  {
    $dbintf->SetScimUserPhones($user['accid'], []);
  }
  else if($operation == "replace")
  {
    $phones = CreateUserPhones(["phonenumbers" => $value, "displayname" => $user["name"]]);
    if(!$dbintf->SetScimUserPhones($user['accid'], $phones))
    {
      throw new ScimException($dbintf->lasterr, 500);
    }
  }
  else if($operation == "add")
  {
		if(isset($value["value"]))
		{
			$value = [$value];
		}
    $phones = CreateUserPhones(["phonenumbers" => $value, "displayname" => $user["name"]]);

    // With empty replace list, so will only add the numbers
    ReplaceScimUserPhones($user["accid"], $phones, []);
  }
}

function PatchAttributePath($patch, $node, &$user)
{
  $value = $patch["value"] ?? null;
  $op = $patch["op"];

  switch(strtolower($node->attributeNames[0]))
  {
    case 'username':
      UpdateUserName($user, $value, $op);
      break;  
    case 'active':
      UpdateActive($user, $value, $op);
      break;
    case 'externalid':
      UpdateExternalID($user, $value, $op);
      break;
    case 'displayname':
      UpdateDisplayName($user, $value, $op);
      break;
    case 'emails':
      UpdateEmails($user, $value, $op);
      break;
    case 'phonenumbers':
      UpdatePhoneNumbers($user, $value, $op);
      break;
    default:
      throw new ScimException("Unsupported patch path: " . implode(".", $node->attributeNames), 400, "invalidPath");
  }
}

function FilterConjuction($node, $compareFunc)
{
  $factors = $node->GetFactors();
	$ret = true;
	foreach($factors as $childNode)
	{
		$ret = $ret && FilterApplies($childNode, $compareFunc);
	}

  return $ret;
}

function FilterDisjuction($node, $compareFunc)
{
  $terms = $node->GetTerms();
	$ret = false;
	foreach($terms as $childNode)
	{
		$ret = $ret || FilterApplies($childNode, $compareFunc);
	}

  return $ret;
}

function FilterNegation($node, $compareFunc)
{
  return !FilterApplies($node->filter, $compareFunc);
}


function FilterComparison($node, $compareFunc)
{
  $attributeName =  strtolower(implode(".", $node->attributePath->attributeNames));
  return $compareFunc($attributeName, $node->operator, $node->compareValue);
}


function FilterApplies($node, $compareFunc)
{
  if($node instanceof Tmilos\ScimFilterParser\Ast\Conjunction)
	{
    return FilterConjuction($node, $compareFunc);
	}
  else if($node instanceof Tmilos\ScimFilterParser\Ast\Disjunction)
  {
    return FilterDisjuction($node, $compareFunc);
  }
  else if($node instanceof Tmilos\ScimFilterParser\Ast\ComparisonExpression)
  {
    return FilterComparison($node, $compareFunc);
  }
  else if($node instanceof Tmilos\ScimFilterParser\Ast\Negation)
  {
    return FilterNegation($node, $compareFunc);
  }
  else
  {
    throw new ScimException("Unsupported filter construct", 400, "invalidFilter");
  }
}

function StringCompare($column, $operator, $value)
{
  switch($operator)
  {
    case "eq":
      return $column == $value;
    case "ne":
      return $column != $value;
    case "co":
      $ix = strpos($column, $value);
      return $ix !== false;
    case "sw":
      return substr_compare($column, $value, 0) === 0;
    case "ew":
      $length = strlen($value);
      return substr_compare($column, $value, -$length) === 0;
    default:
      throw new ScimException("Unsupported comparison operator '$operator' for string type", 400, "invalidFilter");
  }
}

function BoolCompare($column, $operator, $value)
{
  switch($operator)
  {
    case "eq":
      return $column == $value;
    case "ne":
      return $column != $value;
    default:
      throw new ScimException("Unsupported comparison operator '$operator' for boolean type", 400, "invalidFilter");
  }  
}

function UpdateEmailsWithFilter($patch, $filter, $attributePath, &$user)
{
  $extra = StringToArray($user['extra']);
  $email = [
    "email" => $user["email"],
    "primary" => $extra["email_primary"] == "1",
    "type" => $extra["email_type"]
  ];

  $applies = FilterApplies($filter, function($attribute, $operator, $value) use($email)
  {
    switch(strtolower($attribute))
    {
      case "value":
        return StringCompare($email["email"], $operator, $value);
      case "type":
        return StringCompare($email["type"], $operator, $value);
      case "primary":
        return BoolCompare($email["primary"], $operator, $value);
    }
  });

  $operator = strtolower($patch["op"]);

  if(!$applies && $operator != "add")
  {
    throw new ScimException("No records have been selected for replace operation", 400, "invalidPath");
  }
  else if(!$applies && $operator == "add")
  {
    $typeFromFilter = ToSqlFilter($filter, "GetExtraFromFilter")["toBind"]["type"];
  }

  switch(strtolower($patch["op"]))
  {
    case "add":
    case "replace":
      $value = $patch["value"] ?? null;
      
      if(!isset($value))
      {
        throw new ScimException("No value for operation", 400, "invalidValue");
      }

      if(!isset($attributePath))
      {
        GetEmailFromScim($user, ["emails" => $value]);
      }
      else
      {
        switch(strtolower($attributePath->attributeNames[0]))
        {
          case "value":
            $user["email"] = $value;
            if(empty($extra["email_type"]) && !empty($typeFromFilter))
            {
              $extra["email_type"] = $typeFromFilter;
              $user["extra"] = ArrayToString($extra);
            }
            break;
          case "type":
            $extra["email_type"] = $value;
            $user["extra"] = ArrayToString($extra);
            break;
          case "primary":
            $extra["email_primary"] = ToBool($value) ? 1:0;

            if(empty($extra["email_type"]) && !empty($typeFromFilter))
            {
              $extra["email_type"] = $typeFromFilter;
            }

            $user["extra"] = ArrayToString($extra);
            break;
        }
      }
      break;
    case "remove":
      if(!isset($attributePath))
      {
        $user["email"] = "";
        unset($extra["email_primary"]);
        unset($extra["email_type"]);
        $user["extra"] = ArrayToString($extra);
      }
      else
      {
        switch(strtolower($attributePath->attributeNames[0]))
        {
          case "value":
            $user["email"] = "";
            break;
          case "type":
            unset($extra["email_type"]);
            $user["extra"] = ArrayToString($extra);
            break;
          case "primary":
            unset($extra["email_primary"]);
            $user["extra"] = ArrayToString($extra);
            break;
        }
      }
  }
}

function GetExtraFromFilter($node, &$toBind)
{
  $attributeName = strtolower(implode(".", $node->attributePath->attributeNames));
  
  if($node->operator == "eq")
  {
    switch($attributeName)
    {
      case "value":
        $toBind["value"] = $node->compareValue;
        break;
      case "type":
        $toBind["type"] = $node->compareValue;
        break;
      case "primary":
        $toBind["type"] = ToBool($node->compareValue);
        break;
    }
  }
}

function PhoneSqlComparison($node, &$toBind)
{
  $attributeName =  strtolower(implode(".", $node->attributePath->attributeNames));
  switch($attributeName)
  {
    case "value": 
      $validOps = ["eq", "ne", "co", "sw", "ew"];
      
      if(!in_array($node->operator, $validOps))
      {
        throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
      }

      return CreateComparison("cd.callerid", $node->operator, $node->compareValue, $toBind);
    case "type":
      $validOps = ["eq", "ne"];
      
      if(!in_array($node->operator, $validOps))
      {
        throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
      }

      return CreateComparison("s.type", $node->operator, $node->compareValue, $toBind);
    case "primary":
      $validOps = ["eq", "ne"];
      
      if(!in_array($node->operator, $validOps))
      {
        throw new ScimException("Unsupported operator '$node->operator' for attribute '$attributeName'", 400, "invalidFilter");
      }

      return CreateComparison("s.is_primary", $node->operator, ToBool($node->compareValue), $toBind);
    default:
      throw new ScimException("Unsupported filter attribute '$attributeName'", 400, "invalidFilter");
  }
}

function ToContactList($phoneList, $id="contactid")
{
  return array_map(function($phone) use($id)
  {
    return $phone[$id];
  }, $phoneList);
}

function ReplaceScimUserPhones($accid, $newPhones, $replacePhones)
{
  global $dbintf;
  $existingPhones = $dbintf->GetScimUserPhones($accid);

  $existingContacts = ToContactList($existingPhones);
  $newContacts = ToContactList($newPhones);
  $replaceContacts = ToContactList($replacePhones, "userphoneid");

	$AnyPrimary = false;
	foreach($newPhones as $phone)
	{
		if($phone["scim_data"]["is_primary"])
		{
			$AnyPrimary = true;
		}
	}

  foreach($existingPhones as $existing)
  {
		if(in_array($existing['userphoneid'], $replaceContacts)) continue;

		$newPhones[] = ["contactid" => (int) $existing["contactid"],
			"scim_data" => [
				"type" => str_ensure_defined($existing["scim_data"]["type"]),
				"is_primary" => ($AnyPrimary ? false : $existing["scim_data"]["is_primary"])
			]
		];
  }

	$newPhones = array_unique($newPhones, SORT_REGULAR);

  if(!$dbintf->SetScimUserPhones($accid, $newPhones))
  {
    throw new ScimException($dbintf->lasterr, 500);
  }
}

function UpdatePhoneNumbersWithFilter($patch, $astFilter, $attributePath, &$user)
{
  global $dbintf;
  
  $filter = ToSqlFilter($astFilter, "PhoneSqlComparison");
  $selectedPhones = $dbintf->GetScimUserPhones($user['accid'], $filter);
 
  $op = strtolower($patch["op"]);

  if($op == "remove")
  {
    if(isset($attributePath))
    {
      $attributeName =  strtolower(implode(".", $attributePath->attributeNames));
      if($attributeName == "value")
      {
        throw new ScimException("Cannot create empty phonenumber", 500);
      }

      foreach($selectedPhones as $selected)
      {
        $scimData = $selected['scim_data'];
        switch($attributeName)
        {
          case "primary":
            $scimData["is_primary"] = false;
            break;
          case "type":
            $scimData["type"] = null;
            break;
          default:
            throw new ScimException("Unknown subattribute $attributeName", 400, "invalidFilter");
        }

        $dbintf->UpdateScimUserPhone($selected['userphoneid'], $scimData);
      }
    }
    else
    {
      ReplaceScimUserPhones($user['accid'], [], $selectedPhones);
    }
  }
  else if($op == "replace" || $op == "add")
  {
    if(!isset($patch["value"]))
    {
      throw new ScimException("Expected value to be set", 400, "invalidValue");
    }

    if($op == "replace" && count($selectedPhones) == 0)
    {
      throw new ScimException("No records have been selected for replace operation", 400, "invalidPath");
    }
    if($op == "add" && count($selectedPhones) == 0)
    {
      $typeFromFilter = ToSqlFilter($astFilter, "GetExtraFromFilter")["toBind"]["type"];
    }

    if(isset($attributePath))
    {
      $attributeName =  strtolower(implode(".", $attributePath->attributeNames));
      if($attributeName == "value")
      {
        if(isset($selectedPhones[0]))
        {
          $scim_data = $selectedPhones[0]['scim_data'];
        }
        else if(!empty($typeFromFilter))
        {
          $scim_data = 
          [
            "type" => $typeFromFilter,
						"is_primary" => false
          ];
        }
        else
        {
          $scim_data = [];
        }

        $newPhones = CreateUserPhones(
          [ "phonenumbers" => [
              [
                "value" => $patch["value"],
                "type" => $scim_data["type"],
                "primary" => $scim_data["is_primary"]
              ]
            ],
            "displayname" => $user["name"]
          ],
        );

        ReplaceScimUserPhones($user['accid'], $newPhones, $selectedPhones);
      }
      else
      {
        foreach($selectedPhones as $selected)
        {
          $scimData = $selected['scim_data'];
          switch($attributeName)
          {
            case "primary":
              $scimData["is_primary"] = ToBool($patch["value"]);
              break;
            case "type":
              $scimData["type"] = $patch["value"];
              break;
            default:
              throw new ScimException("Unknown subattribute $attributeName", 400, "invalidFilter");
          }
  
          $dbintf->UpdateScimUserPhone($selected['userphoneid'], $scimData);

					if($attributeName == "primary" && ToBool($patch["value"]) == true)
					{
						$userPhones = $dbintf->GetScimUserPhones($user['accid']);
						foreach($userPhones as $phone)
						{
							if($phone['userphoneid'] == $selected['userphoneid']) continue;
							if($phone['scim_data']['is_primary'] == false) continue;
							$dbintf->UpdateScimUserPhone($phone['userphoneid'], ["is_primary" => false]);
						}
					}
        }
      }
    }
    else
    {
      if(!is_array($patch["value"]) && !isset($patch["value"]["value"]))
      {
        throw new ScimException("Expected value to be an object", 400, "invalidValue");
      }

			$value = $patch["value"];
			if(isset($value["value"]))
			{
				$value = [$value];
			}
      $newPhones = CreateUserPhones([
        "phonenumbers" => $value,
        "displayname" => $user["name"]
      ]);

      ReplaceScimUserPhones($user['accid'], $newPhones, $selectedPhones);
    }
  }
}


function PatchValuePath($patch, $valuePath, $attributePath, &$user)
{
  $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\ValuePath::class, 'attributePath');
  $reflectionProperty->setAccessible(true);
  $innerAttributePath = $reflectionProperty->getValue($valuePath);

  $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\ValuePath::class, 'filter');
  $reflectionProperty->setAccessible(true);
  $filter = $reflectionProperty->getValue($valuePath);

  switch(strtolower($innerAttributePath->attributeNames[0]))
  {
    case 'emails':
      UpdateEmailsWithFilter($patch, $filter, $attributePath, $user);
      break;
    case 'phonenumbers':
      UpdatePhoneNumbersWithFilter($patch, $filter, $attributePath, $user);
      break;
    default:
      throw new ScimException("Unsupported patch path: " . implode(".", $innerAttributePath->attributeNames), 400, "invalidPath");
  }
}

function PatchWithoutPath($value, &$user)
{
  global $dbintf;
  if(!isset($value))
  {
    throw new ScimException("Expected a value in the patch operation", 400, "invalidValue");
  }

  $value = array_change_key_case($value, CASE_LOWER);

  if(isset($value["username"]))
  {
    if($value['username'] != $user['usr'] && $dbintf->UsrExists($value['username']))
    {
      throw new ScimException("A user with the username exists already", 409);
    }

    $user['usr'] = $value["username"];
  }

  if(isset($value["displayname"]))
  {
    $user['name'] = $value["displayname"];
  }

  if(isset($value["active"]))
  {
    $user['enabled'] = ToBool($value["active"]) ? 1:0;
  }

  if(isset($value["externalid"]))
  {
    $user["externalid"] = $value["externalid"];
  }
}

function DoPatchOperation($patch, &$user)
{
  $patch = array_change_key_case($patch, CASE_LOWER);

  if(isset($patch['path']))
  {
    $parser = new Parser(Mode::PATH());
		$node = $parser->parse($patch['path']);

    $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\Path::class, 'attributePath');
    $reflectionProperty->setAccessible(true);
    $attributePath = $reflectionProperty->getValue($node);

    $reflectionProperty = new \ReflectionProperty(Tmilos\ScimFilterParser\Ast\Path::class, 'valuePath');
    $reflectionProperty->setAccessible(true);
    $valuePath = $reflectionProperty->getValue($node);

    if(!isset($valuePath))
    {
      PatchAttributePath($patch, $attributePath, $user);
    }
    else
    {
      PatchValuePath($patch, $valuePath, $attributePath, $user);
    }
  }
  else
  {
    if(strtolower($patch["op"]) == "remove")
    {
      throw new ScimException("No target path has been specified for remove operation", 400, "noTarget");
    }

    PatchWithoutPath($patch["value"] ?? null, $user);
  }
}

function PatchUser($id, $patch, $parentInfo)
{
  global $dbintf;

  if(!isset($id))
	{
		throw new ScimException("No user id specified", 400, "invalidValue");
	}

	if(!isset($patch))
	{
		throw new ScimException("No request body available", 400, "invalidSyntax");
	}

  $patch = array_change_key_case($patch, CASE_LOWER);

  if(!CheckPatchSchema($patch))
  {
    throw new ScimException("Unsupported schema", 400, "invalidSyntax");
  }

  $user = $dbintf->GetScimUser($id, $parentInfo['id']);

  if($user === false)
  {
    throw new ScimException("User not found", 404);
  }

  $dbintf->StartTransaction();
  $oldExtra = StringToArray($user["extra"]);

  foreach($patch["operations"] as $operation)
  {
    DoPatchOperation($operation, $user);
  }

  $dbuser = 
  [
    "usr" => $user["usr"],
    "name" => $user["name"],
    "enabled" => $user["enabled"],
  ];

  $newExtra = StringToArray($user["extra"]);

  if($oldExtra["email_primary"] !== "1" || $newExtra["email_primary"] === "1")
  {
    $dbuser["email"] = $user["email"];
    $dbuser["extra"] = $user["extra"];
  }

  $time = new DateTime("now", new DateTimeZone("UTC"));

  $scimData =
  [
    "updated" => $time->format(DateTimeInterface::ATOM),
    "externalid" => $user["externalid"]
  ];

  if(!$dbintf->EditUserAccount($id, $dbuser))
	{
		throw new ScimException($dbintf->lasterr, 500);
	}

  if(!$dbintf->EditScimData($user['accid'], $scimData))
	{
		throw new ScimException("Could not write updated user information to database", 500);
	}

  $dbintf->CommitTransaction();
  EnginesSettingsCalcApply();
}

function QueryParamLog()
{
  $logs = [];
  if(isset($_GET['filter']))
  {
    $logs[] = 'filter: ' . $_GET['filter'];
  }

  if(isset($_GET['attributes']))
  {
    $logs[] = 'attributes: ' . $_GET['attributes'];
  }

  if(isset($_GET['excludedAttributes']))
  {
    $logs[] = 'excludedAttributes: ' . $_GET['excludedAttributes'];
  }

  return 'QueryParams: ' . implode(",", $logs);
}

function HandleRequest()
{
	if(!isset($_GET['group']))
	{
		throw new ScimException("Not found", 404);
	}

	global $requestID;
	$requestID = uniqid();
	$parentGroup = $_GET['group'];

  global $dbintf;
  InitDB();

	$parentInfo = $dbintf->GetSingleGroupInfoByName($parentGroup);
	if($parentInfo === false)
	{
		throw new ScimException("Not found", 404);
	}

  if(!IsAuthorized($parentInfo))
  {
    throw new ScimException("Unauthorized", 401);
  }

	$dbintf->SetAdmin();

  $method = $_SERVER['REQUEST_METHOD'];

  switch($_SERVER['REQUEST_METHOD'])
  {   
    case 'GET':
			if(isset($_GET['id']))
			{
        scimlg("SCIM user get with ID: " . $_GET['id'] . ", " . QueryParamLog());
      	$result = GetUser($_GET['id'], $parentInfo);
			}
			else
			{
        scimlg("SCIM user get without ID, " . QueryParamLog());
				$result = GetUsers($parentInfo);
			}

      $response = json_encode($result);
      scimlg("SCIM User GET response: " . $response);
      header('Content-Type: application/scim+json');
      echo $response;

      break;
    case 'POST':
      $input = file_get_contents("php://input");
      $body = json_decode($input, true);
      scimlg("SCIM User POST request: " . json_encode($body) . ", " . QueryParamLog());
      $result = NewUser($body, $parentInfo);
      $response = json_encode($result);
      scimlg("SCIM User POST response: " . $response);
      http_response_code(201);
      header('Content-Type: application/scim+json');
      echo $response;
      break;
    case 'PATCH':
      $input = file_get_contents("php://input");
      $body = json_decode($input, true);
      scimlg("SCIM User " .  $_GET['id'] . " PATCH request: " . json_encode($body) . ", " . QueryParamLog());
      PatchUser($_GET['id'], $body, $parentInfo);
      http_response_code(200);
      break;
		case 'PUT':
			$input = file_get_contents("php://input");
			$body = json_decode($input, true);
      scimlg("SCIM User " . $_GET['id'] . " PUT request: " . json_encode($body) . ", " . QueryParamLog());
			ReplaceUser($_GET['id'], $body, $parentInfo);
			http_response_code(200);
			break;
    case 'DELETE':
      scimlg("SCIM User " . $_GET['id'] . " DELETE request" . ", " . QueryParamLog());
      DeleteUser($_GET['id'], $parentInfo);
      http_response_code(204);
      break;
    default:
      throw new ScimException("This method is not supported: $method", 501);
  }
}

function CancelDB()
{
  global $dbintf;
  if(isset($dbintf))
  {
    while($dbintf->transactionLevel > 0)
    {
      $dbintf->CancelTransaction();
    }
  }
}

try
{
  HandleRequest();
}
catch(ScimException $exception)
{
  $result = [
    "schemas" => [ERROR_SCHEMA],
    "detail" => $exception->getMessage(),
    "status" => $exception->getCode()
  ];

  if(isset($exception->scimType))
  {
    $result["scimType"] = $exception->scimType;
  }

  $code = $exception->getCode();
  $response = json_encode($result);
  scimlg('Scim exception on user: ' . $response);

  header('Content-Type: application/scim+json');
  http_response_code($code);
  echo $response;
}
catch(Tmilos\ScimFilterParser\Error\FilterException | Tmilos\Lexer\Error\UnknownTokenException $exception)
{
  $result = [
    "schemas" => [ERROR_SCHEMA],
    "detail" => "Filter syntax error",
    "scimType" => "invalidFilter",
    "status" => 400
  ];

  $response = json_encode($result);
  scimlg('Scim filter parse exception on user: ' . $response);

  header('Content-Type: application/scim+json');
  http_response_code(400);
  echo $response;
}
catch(Exception $exception)
{
	$result = [
    "schemas" => [ERROR_SCHEMA],
    "detail" => "An internal error occured",
    "status" => 500
  ];

	$response = json_encode($result);
  scimlg('Exception on user: ' . $exception->getMessage());
	header('Content-Type: application/scim+json');
  http_response_code(500);
  echo $response;
}
finally
{
  CancelDB();
}

?>